from numbers import Number
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.service import StringHelper, LogHelper
from python_helper.api.src.helper import ObjectHelperHelper

GENERATOR_CLASS_NAME = 'generator'
UNKNOWN_OBJECT_CLASS_NAME = c.UNKNOWN.lower()

METADATA_NAME = 'metadata'

NATIVE_CLASSES = (
    bool,
    int,
    str,
    float,
    bytes,
    type(ObjectHelperHelper.generatorInstance())
)

COLLECTION_CLASSES = (
    list,
    dict,
    tuple,
    set
)

def equals(
    expected,
    toAssert,
    ignoreKeyList = None,
    ignoreCharactereList = None,
    visitedIdInstances = None,
    muteLogs = True
) :
    if isNone(expected) or isNone(toAssert) :
        return expected is None and toAssert is None
    if isNativeClass(type(expected)) :
        return expected == toAssert
    if isNone(visitedIdInstances) :
        visitedIdInstances = []
    if isDictionary(expected) and isDictionary(toAssert) :
        innerIgnoreCharactereList = [c.SPACE]
        if isNotNone(ignoreCharactereList) :
            innerIgnoreCharactereList += ignoreCharactereList
        filteredResponse = StringHelper.filterJson(
            str(sortIt(filterIgnoreKeyList(expected,ignoreKeyList))),
            extraCharacterList=innerIgnoreCharactereList
        )
        filteredExpectedResponse = StringHelper.filterJson(
            str(sortIt(filterIgnoreKeyList(toAssert,ignoreKeyList))),
            extraCharacterList=innerIgnoreCharactereList
        )
        return filteredResponse == filteredExpectedResponse
    elif isCollection(expected) and isCollection(toAssert) :
        areEquals = True
        try :
            for a, b in zip(expected, toAssert) :
                areEquals = equals(
                    a,
                    b,
                    ignoreKeyList = ignoreKeyList,
                    ignoreCharactereList = ignoreCharactereList,
                    visitedIdInstances = visitedIdInstances,
                    muteLogs = muteLogs
                )
                if not areEquals :
                    break
            return areEquals
        except Exception as exception :
            areEquals = False
            LogHelper.log(equals, f'Different arguments in {expected} and {toAssert}. Returning "{areEquals}" by default', exception=exception)
    else :
        if isNotNone(toAssert) and id(toAssert) not in visitedIdInstances :
            areEquals = True
            try :
                if not muteLogs :
                    LogHelper.prettyPython(equals, f'expected', expected, logLevel = LogHelper.DEBUG, condition=not muteLogs)
                    LogHelper.prettyPython(equals, f'toAssert', toAssert, logLevel = LogHelper.DEBUG, condition=not muteLogs)
                areEquals = True and ObjectHelperHelper.leftEqual(expected, toAssert, visitedIdInstances, muteLogs=muteLogs) and ObjectHelperHelper.leftEqual(toAssert, expected, visitedIdInstances, muteLogs=muteLogs)
            except Exception as exception :
                areEquals = False
                LogHelper.log(equals, f'Different arguments in {expected} and {toAssert}. Returning "{areEquals}" by default', exception=exception)
            visitedIdInstances.append(id(toAssert))
            return areEquals
        else :
             return True

def sortIt(thing) :
    if isDictionary(thing) :
        sortedDictionary = {}
        for key in getSortedCollection(thing) :
            sortedDictionary[key] = sortIt(thing[key])
        return sortedDictionary
    elif isCollection(thing) :
        newCollection = []
        for innerValue in thing :
            newCollection.append(sortIt(innerValue))
        return getSortedCollection(newCollection)
    else :
        return thing

def getSortedCollection(thing) :
    return thing if (
        isNotCollection(thing) or isEmpty(thing)
    ) or (
        isNotDictionary(thing) and isNotSet(thing) and isDictionary(thing[0])
    ) else sorted(
        thing,
        key=lambda x: (
            x is not None, c.NOTHING if isinstance(x, Number) else type(x).__name__, x
        )
    )

def filterIgnoreKeyList(objectAsDictionary,ignoreKeyList):
    if isDictionary(objectAsDictionary) and isNotNone(ignoreKeyList) :
        filteredObjectAsDict = {}
        for key, value in objectAsDictionary.items() :
            if key not in ignoreKeyList :
                if isDictionary(value) :
                    filteredObjectAsDict[key] = filterIgnoreKeyList(value,ignoreKeyList)
                else :
                    filteredObjectAsDict[key] = objectAsDictionary[key]
        return filteredObjectAsDict
    return objectAsDictionary

def isEmpty(thing) :
    return StringHelper.isBlank(thing) if isinstance(thing, str) else isNone(thing) or isEmptyCollection(thing)

def isNotEmpty(thing) :
    return not isEmpty(thing)

def isEmptyCollection(thing) :
    return isCollection(thing) and 0 == len(thing)

def isNotEmptyCollection(thing) :
    return isCollection(thing) and 0 < len(thing)

def isList(thing) :
    return isinstance(thing, list)

def isNotList(thing) :
    return not isList(thing)

def isSet(thing) :
    return isinstance(thing, set)

def isNotSet(thing) :
    return not isSet(thing)

def isTuple(thing) :
    return isinstance(thing, tuple)

def isNotTuple(thing) :
    return not isTuple(thing)

def isDictionary(thing) :
    return isinstance(thing, dict)

def isNotDictionary(thing) :
    return not isDictionary(thing)

def isDictionaryClass(thingClass) :
    return dict == thingClass

def isNotDictionaryClass(thingClass) :
    return not isDictionaryClass(thingClass)

def isNone(instance) :
    return instance is None

def isNotNone(instance) :
    return not isNone(instance)

def isNativeClass(instanceClass) :
    return isNotNone(instanceClass) and instanceClass in NATIVE_CLASSES

def isNotNativeClass(instanceClass) :
    return not isNativeClass(instanceClass)

def isNativeClassInstance(instance) :
    return isNotNone(instance) and isNativeClass(instance.__class__)

def isNotNativeClassIsntance(instance) :
    return not isNativeClassInstance(instance)

def isCollection(instance) :
    return isNotNone(instance) and True in {isinstance(instance, c) for c in COLLECTION_CLASSES}

def isNotCollection(instance) :
    return not isCollection(instance)

def isNeitherNoneNorBlank(thing) :
    return isNotNone(thing) and StringHelper.isNotBlank(str(thing))

def isNoneOrBlank(thing) :
    return isNone(thing) or StringHelper.isBlank(str(thing))



from python_helper.api.src.annotation import TestAnnotation
from python_helper.api.src.annotation.TestAnnotation import Test
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.service import SettingHelper, LogHelper, ReflectionHelper, ObjectHelper, StringHelper
import time

PY = 'py'
DOT_PY = f'{c.DOT}{PY}'
TEST_SUFIX_NAME = 'Test'
TEST_DOT_PY = f'{TEST_SUFIX_NAME}{DOT_PY}'
TEST_PACKAGE = TEST_SUFIX_NAME.lower()

GLOBALS_INSTANCE_LIST = []
RESULTS = {}
TEST_KWARGS = {
    'argsOfCallBefore' : [GLOBALS_INSTANCE_LIST],
    'argsOfCallAfter' : [GLOBALS_INSTANCE_LIST],
    'returns' : RESULTS,
    'logResult' : False
}

def getRaisedException(callableThing, *args, logGetRaisedExceptionIfAny=False, **kwargs):
    try:
        callableThing(*args, **kwargs)
    except Exception as exception:
        if logGetRaisedExceptionIfAny:
            LogHelper.error(getRaisedException, 'Exception', exception)
        return exception

def getUnitTest(inspectGlobals, globalsInstance) :
    @Test(**getGlobalsTestKwargs(inspectGlobals, globalsInstance))
    def unitTest(testModule, testName, data, testReturns, logResult) :
        discountTimeEnd = time.time()
        unitTestException = None
        if logResult :
            LogHelper.test(unitTest, f'{testName}{c.DOT}{data[1]} test started')
        moduleTestStartTime = time.time()
        try :
            testReturns[data[1]] = data[0]()
            LogHelper.printSuccess(f'{data[0].__module__}{c.DOT}{data[0].__name__} succeed', condition=logResult, newLine=False, margin=False)
        except Exception as exception :
            LogHelper.printError(f'{data[0].__module__}{c.DOT}{data[0].__name__} failed', condition=True, newLine=False, margin=False, exception=exception)
            unitTestException = exception
        if logResult :
            LogHelper.test(unitTest, f'{testName}{c.DOT}{data[1]} test completed in {time.time() - moduleTestStartTime} seconds')
        if ObjectHelper.isNotNone(unitTestException) :
            raise unitTestException
        someDidRun = True
        return testName, data, testReturns, someDidRun, logResult, discountTimeEnd
    return unitTest

def getUnitTestBatch(inspectGlobals, globalsInstance) :
    @Test(**getGlobalsTestKwargs(inspectGlobals, globalsInstance))
    def unitTestBatch(testModule, data) :
        discountTimeEnd = time.time()
        data[0]()
        return discountTimeEnd
    return unitTestBatch

def getModuleTest(inspectGlobals, logResult, globalsInstance) :
    @Test(**getGlobalsTestKwargs(inspectGlobals, globalsInstance))
    def tddModule(testName, testModule, dataList, times, runSpecificTests, testsToRun, allShouldRun, someShouldRun, logResult) :
        import globals
        tddModuleGlobalsInstance = globals.newGlobalsInstance(testModule.__file__
            , successStatus = globalsInstance.successStatus
            , errorStatus = globalsInstance.errorStatus
            , failureStatus = globalsInstance.failureStatus
            , settingStatus = globalsInstance.settingStatus or inspectGlobals
            , statusStatus = globalsInstance.statusStatus
            , infoStatus = globalsInstance.infoStatus
            , debugStatus = globalsInstance.debugStatus
            , warningStatus = globalsInstance.warningStatus
            , wrapperStatus = globalsInstance.wrapperStatus
            , testStatus = globalsInstance.testStatus
            , logStatus = globalsInstance.logStatus
            , logsWithColors = globalsInstance.logsWithColors or LogHelper.colorsEnabled()
        )
        LogHelper.test(tddModule, f'{testName} started')
        testReturns = {}
        testTime = 0
        for data in dataList :
            testMustRun = isTestToRun(testModule, data, runSpecificTests, testsToRun)
            LogHelper.prettyPython(tddModule, f'test attribute or method', data[0], logLevel=LogHelper.TEST)
            LogHelper.prettyPython(tddModule, f'isTestToRun', testMustRun, logLevel=LogHelper.TEST)
            if testMustRun :
                for index in range(times) :
                    testTimeStart = time.time()
                    if times - 1 == index :
                        runnableUnitTest = getUnitTest(inspectGlobals, globalsInstance)
                        testName, data, testReturns, someDidRun, logResult, discountTimeEnd = runnableUnitTest(testModule, testName, data, testReturns, logResult)
                    else :
                        runnableUnitTestBatch = getUnitTestBatch(inspectGlobals, globalsInstance)
                        discountTimeEnd = runnableUnitTestBatch(testModule, data)
                    testTime += time.time() - discountTimeEnd
            else :
                allDidRun = False
        if not allShouldRun == allDidRun and someShouldRun == someDidRun :
            amountAlthought = getAmmount(allShouldRun, someShouldRun)
            shouldOrShouldnt = getShouldOrShouldntAsString(allShouldRun, someShouldRun)
            amountAfterAlthought = getAmmount(allDidRun, someDidRun)
            didOrDidnt = getSomeDidOrDidntAsString(allDidRun, someDidRun)
            errorMessage = f'Inconsistenc{StringHelper.getS(allShouldRun or someShouldRun, es=True)} detected. Although {amountAlthought} test{StringHelper.getS(allShouldRun or someShouldRun)} {shouldOrShouldnt} run, {amountAfterAlthought} {didOrDidnt} run'
            exception = Exception(errorMessage)
            raise exception
        return allDidRun, someDidRun, testTime, testReturns
    return tddModule

def runModuleTests(testName, runnableTddModule, times, runSpecificTests, testsToRun, logResult, globalsInstance) :
    import globals
    testModule = globals.importModule(testName, required=True)
    # if ObjectHelper.isNone(testModule):
    #     errorMessage = f'Not possible to import {testName}'
    #     exception = Exception(f'Not possible to import {testName}')
    #     LogHelper.error(runModuleTests, f'{errorMessage}. Please check warnings for more details', exception)
    #     raise exception
    dataList = ReflectionHelper.getAttributeDataList(testModule)
    LogHelper.prettyPython(runnableTddModule, f'{ReflectionHelper.getName(testModule)} tests loaded', dataList, logLevel=LogHelper.TEST)
    allShouldRun, someShouldRun, allDidRun, someDidRun = getRunStatus(testModule, dataList, runSpecificTests, testsToRun)
    testReturns = {}
    testTime = 0
    totalTestTimeStart = time.time()
    if someShouldRun :
        defaultMessage = f'{testName}{StringHelper.getS(allShouldRun)}'
        methodReturnException = None
        LogHelper.test(runModuleTests, f'{defaultMessage} started')
        try :
            allDidRun, someDidRun, testTime, testReturns = runnableTddModule(testName, testModule, dataList, times, runSpecificTests, testsToRun, allShouldRun, someShouldRun, logResult)
            LogHelper.printSuccess(f'{defaultMessage} succeed. {getTestRuntimeInfo(times, testTime, time.time() - totalTestTimeStart)}', condition=logResult, newLine=True, margin=False)
        except Exception as exception :
            methodReturnException = exception
            LogHelper.printError(f'{defaultMessage} failed', condition=True, newLine=True, margin=False, exception=methodReturnException)
        exceptionRaised = ObjectHelper.isNotNone(methodReturnException)
        defaultMessage = f'{testName}{StringHelper.getS(not exceptionRaised and allDidRun)}'
        if exceptionRaised :
            raise methodReturnException
    return allDidRun, someDidRun, testTime, testReturns

def getRunStatus(testModule, dataList, runSpecificTests, testsToRun) :
    allDidRun = True
    someDidRun = False
    allShouldRun = True
    someShouldRun = False
    for data in dataList :
        if isTestToRun(testModule, data, runSpecificTests, testsToRun) :
            someShouldRun = True
        else :
            allShouldRun = False
    allShouldRun, someShouldRun, allDidRun, someDidRun
    return allShouldRun, someShouldRun, allDidRun, someDidRun

def getAmmount(allShouldRun, someShouldRun) :
    if allShouldRun :
        return 'all'
    elif someShouldRun :
        return 'some'
    else :
        return 'none'

def getShouldOrShouldntAsString(allShouldRun, someShouldRun) :
    return f'should' if allShouldRun or someShouldRun else f'shouldn{c.SINGLE_QUOTE}t'

def getSomeDidOrDidntAsString(allDidRun, someDidRun) :
    if allDidRun :
        return 'all'
    elif someDidRun :
        return 'some'
    else :
        return 'none'

def getSomeDidOrDidntAsString(allDidRun, someDidRun) :
    return 'did' if allDidRun or someDidRun else f'didn{c.SINGLE_QUOTE}t'

def run(
    filePath,
    runOnly = None,
    ignore=None,
    times = 1,
    successStatus = True,
    errorStatus = True,
    settingStatus = False,
    statusStatus = True,
    infoStatus = True,
    debugStatus = True,
    warningStatus = True,
    failureStatus = True,
    wrapperStatus = False,
    testStatus = False,
    logStatus = False,
    logsWithColors = False,
    inspectGlobals = False,
    logResult = True
) :
    import globals
    globalsInstance = globals.newGlobalsInstance(filePath
        , successStatus = successStatus
        , errorStatus = errorStatus
        , settingStatus = settingStatus or inspectGlobals
        , statusStatus = statusStatus
        , infoStatus = infoStatus
        , debugStatus = debugStatus
        , warningStatus = warningStatus
        , failureStatus = failureStatus
        , wrapperStatus = wrapperStatus
        , testStatus = testStatus
        , logStatus = logStatus
        , logsWithColors = logsWithColors or LogHelper.colorsEnabled()
    )
    testModuleNames, testsToRun, runSpecificTests = getTestModuleNames(runOnly, ignore, globalsInstance)
    returns = {}
    totalTestTimeStart = time.time()
    testTime = 0
    for testModuleName in testModuleNames :
        runnableTddModule = getModuleTest(inspectGlobals, logResult, globalsInstance)
        allDidRun, didRun, moduleTestTime, testReturns = runModuleTests(testModuleName, runnableTddModule, times, runSpecificTests, testsToRun, logResult, globalsInstance)
        returns[testModuleName] = testReturns
        testTime += moduleTestTime
    totalTestTime = time.time() - totalTestTimeStart
    if logResult :
        LogHelper.success(run, f'{globalsInstance.apiName} tests completed. {getTestRuntimeInfo(times, testTime, totalTestTime)}')
    return returns

def getTestModuleNames(runOnly, ignore, globalsInstance) :
    testsToRun = [] if ObjectHelper.isNone(runOnly) or ObjectHelper.isNotCollection(runOnly) else runOnly
    testsToIgnore = [] if ObjectHelper.isNone(ignore) or ObjectHelper.isNotCollection(ignore) else ignore
    runSpecificTests = ObjectHelper.isNotEmpty(testsToRun)
    LogHelper.prettyPython(getTestModuleNames, f'runSpecificTests: {runSpecificTests}, testsToRun', testsToRun, logLevel=LogHelper.TEST)
    testModuleNames = []
    if ObjectHelper.isEmpty(testsToRun) :
        testQueryTree = SettingHelper.querySetting(TEST_PACKAGE, globalsInstance.apiTree)
        LogHelper.prettyPython(getTestModuleNames, 'Test query tree', testQueryTree, logLevel=LogHelper.TEST)
        testModuleNames += getTestModuleNamesFromQuerryTree(testQueryTree, runOnly, ignore)
    else :
        for testName in testsToIgnore :
            if testName in testsToRun :
                testModuleNames, testsToRun, runSpecificTests .remove(testName)
        for testName in testsToRun :
            testNameSplitted = testName.split(c.DOT)
            testModuleName = c.NOTHING
            if 2 == len(testNameSplitted) :
                testModuleName = testNameSplitted[-2]
            if testModuleName not in testModuleNames and StringHelper.isNotBlank(testName) :
                testModuleNames.append(testModuleName)
    LogHelper.prettyPython(getTestModuleNames, 'Test module names', testModuleNames, logLevel=LogHelper.TEST)
    return testModuleNames, testsToRun, runSpecificTests

def getTestModuleNamesFromQuerryTree(testQueryTree, runOnly, ignore) :
    testModuleNames = []
    for queryResultKey, queryResult in testQueryTree.items() :
        LogHelper.prettyPython(getTestModuleNamesFromQuerryTree, 'Query result', queryResult, logLevel=LogHelper.TEST)
        updateTestModuleNames(testModuleNames, queryResult, runOnly, ignore)
    return testModuleNames

def updateTestModuleNames(testModuleNames, queryResult, runOnly, ignore) :
    if not c.NOTHING == queryResult :
        for key, value in queryResult.items() :
            if c.NOTHING == value and key.endswith(TEST_DOT_PY) :
                testModuleName = key[:-len(DOT_PY)]
                if ObjectHelper.isNone(runOnly) or ObjectHelper.isNotNone(runOnly) and testModuleName in runOnly :
                    testModuleNames.append(testModuleName)
            else :
                updateTestModuleNames(testModuleNames, value, runOnly, ignore)

def isTestToRun(testModule, attributeData, runSpecificTests, testsToRun) :
    return (
            not runSpecificTests or
            (runSpecificTests and f'{ReflectionHelper.getName(testModule)}{c.DOT}{attributeData[1]}' in testsToRun)
        ) and ReflectionHelper.getAttributeOrMethod(attributeData[0], TestAnnotation.IS_TEST_METHOD, muteLogs=True)

def getTestRuntimeInfo(times, testTime, totalTestTime) :
    testRuntimeInfo = None
    try :
        testRuntimeInfo = f'It {StringHelper.getToBe(True, singular=1==times, tense=StringHelper.PAST)} {times} test run{StringHelper.getS(times > 1)} in {testTime} seconds. Total test time: {totalTestTime} seconds'
    except Exception as exception :
        LogHelper.warning(getTestRuntimeInfo, 'Not possible do get test runtime info', exception=exception)
        testRuntimeInfo = c.NOTHING
    return testRuntimeInfo

def getTestEnvironmentVariables(globalsInstance) :
    return {
        LogHelper.LOG : globalsInstance.logStatus,
        LogHelper.INFO : globalsInstance.infoStatus,
        LogHelper.STATUS : globalsInstance.statusStatus,
        LogHelper.SUCCESS : globalsInstance.successStatus,
        LogHelper.SETTING : globalsInstance.settingStatus,
        LogHelper.DEBUG : globalsInstance.debugStatus,
        LogHelper.WARNING : globalsInstance.warningStatus,
        LogHelper.FAILURE : globalsInstance.failureStatus,
        LogHelper.WRAPPER : globalsInstance.wrapperStatus,
        LogHelper.ERROR : globalsInstance.errorStatus,
        LogHelper.TEST : globalsInstance.testStatus,
        LogHelper.ENABLE_LOGS_WITH_COLORS: globalsInstance.logsWithColors
    }

def getGlobalsTestKwargs(inspectGlobals, globalsInstance) :
    import globals
    TEST_KWARGS['callBefore'] = globals.runBeforeTest
    TEST_KWARGS['callAfter'] = globals.runAfterTest
    TEST_KWARGS['inspectGlobals'] = inspectGlobals
    TEST_KWARGS['kwargsOfCallBefore'] = {
        'muteLogs' : not inspectGlobals,
        'logLevel' : LogHelper.DEBUG
    }
    TEST_KWARGS['kwargsOfCallAfter'] = {
        'muteLogs' : not inspectGlobals,
        'logLevel' : LogHelper.DEBUG
    }
    return {
        'environmentVariables' : getTestEnvironmentVariables(globalsInstance),
        **TEST_KWARGS
    }



from python_helper.api.src.service import LogHelper

LOG =       '[LOG    ] '
INFO =      '[INFO   ] '
STATUS =    '[STATUS ] '
SUCCESS =   '[SUCCESS] '
SETTING =   '[SETTING] '
DEBUG =     '[DEBUG  ] '
WARNING =   '[WARNING] '
FAILURE =   '[FAILURE] '
WRAPPER =   '[WRAPPER] '
ERROR =     '[ERROR  ] '
TEST =      '[TEST   ] '

BACK_SLASH = '\\'
NEW_LINE = '\n'
BAR_N = f'{BACK_SLASH}n'
NOTHING = ''
BLANK = ''
SPACE = ' '
SINGLE_QUOTE = "'"
DOUBLE_QUOTE = '"'
TRIPLE_SINGLE_QUOTE = f"{3*SINGLE_QUOTE}"
TRIPLE_DOUBLE_QUOTE = f'{3*DOUBLE_QUOTE}'

PIPE = '|'
BACK_SLASH_SINGLE_QUOTE = "'\'"
BACK_SLASH_DOUBLE_QUOTE = '\"'
SLASH = '/'
FOWARD_SLASH = '/'
DOUBLE_SLASH = f'{2*FOWARD_SLASH}'
HASH_TAG = '#'
DASH = '-'
SPACE_DASH_SPACE = f'{SPACE}{DASH}{SPACE}'
UNDERSCORE = '_'
DOUBLE_UNDERSCORE = f'{2 * UNDERSCORE}'
DOLLAR = '$'
ARROBA = '@'
AND = '&'
PLUS = '+'
EQUALS = '='
PERCENT = '%'
ASTERISK = '*'

DOT = '.'
DOT_SPACE = f'{DOT}{SPACE}'
COMA = ','
COMA_SPACE = f'{COMA}{SPACE}'
COLON = ':'
COLON_SPACE = f'{COLON}{SPACE}'
SPACE_COLON_SPACE = f'{SPACE}{COLON}{SPACE}'
SEMI_COLON = ';'
SEMI_COLON_SPACE = f'{SEMI_COLON}{SPACE}'
QUESTION_MARK = '?'
QUESTION_MARK_SPACE = f'{QUESTION_MARK}{SPACE}'
EXCLAMATION_MARK = '!'
EXCLAMATION_MARK_SPACE = f'{EXCLAMATION_MARK}{SPACE}'

LOG_CAUSE = f'Cause{COLON_SPACE}'
DOT_SPACE_CAUSE = f'{DOT_SPACE}{LOG_CAUSE}'

TAB_UNITS = 4
TAB = TAB_UNITS * SPACE
SYSTEM_TAB = '\t'

NONE = 'None'

TRUE = 'True'
FALSE = 'False'

BOOLEAN = 'bool'
STRING = 'str'
INTEGER = 'int'
TUPLE = 'tuple'
LIST = 'list'
DICT = 'dict'

OPEN_TUPLE = '('
CLOSE_TUPLE = ')'
OPEN_LIST = '['
CLOSE_LIST = ']'
OPEN_DICTIONARY = '{'
CLOSE_DICTIONARY = '}'
OPEN_SET = '{'
CLOSE_SET = '}'
LESSER = '<'
BIGGER = '>'

UNKNOWN = 'Unknown'

TYPE_TYPE = 'type'
TYPE_MODULE = 'module'
TYPE_CLASS = 'class'
TYPE_METHOD = 'method'
TYPE_FUNCTION = 'function'
TYPE_BUILTIN_FUNCTION_OR_METHOD = 'builtin_function_or_method'
TYPE_BOOLEAN = 'bool'
TYPE_STRING = 'str'
TYPE_INTEGER = 'int'
TYPE_FLOAT = 'float'
TYPE_TUPLE = 'tuple'
TYPE_LIST = 'list'
TYPE_DICT = 'dict'
TYPE_SET = 'set'

NATIVE_TYPES = [
    TYPE_TYPE,
    TYPE_MODULE,
    TYPE_CLASS,
    TYPE_METHOD,
    TYPE_FUNCTION,
    TYPE_BUILTIN_FUNCTION_OR_METHOD,
    TYPE_BOOLEAN,
    TYPE_STRING,
    TYPE_INTEGER,
    TYPE_FLOAT,
    TYPE_TUPLE,
    TYPE_LIST,
    TYPE_DICT,
    TYPE_SET
]

NULL_VALUE = 'null'
TRUE_VALUE = 'true'
FALSE_VALUE = 'false'

ENCODING = 'utf-8'
UTF_8 = 'utf-8'
OVERRIDE = 'w+'
READ = 'r'

CSI = '\x1B['
BRIGHT_DIGIT = 1
DARK_DIGIT = 0
DEFAULT_COLOR = CSI + '0m'

DARK_BLACK = CSI + f'{DARK_DIGIT};30m'
BRIGHT_BLACK = CSI + f'{BRIGHT_DIGIT};30m'

DARK_RED = CSI + f'{DARK_DIGIT};31m'
BRIGHT_RED = CSI + f'{BRIGHT_DIGIT};31m'

DARK_GREEN = CSI + f'{DARK_DIGIT};32m'
BRIGHT_GREEN = CSI + f'{BRIGHT_DIGIT};32m'

DARK_YELLOW = CSI + f'{DARK_DIGIT};33m'
BRIGHT_YELLOW = CSI + f'{BRIGHT_DIGIT};33m'

DARK_BLUE = CSI + f'{DARK_DIGIT};34m'
BRIGHT_BLUE = CSI + f'{BRIGHT_DIGIT};34m'

DARK_MAGENTA = CSI + f'{DARK_DIGIT};35m'
BRIGHT_MAGENTA = CSI + f'{BRIGHT_DIGIT};35m'

DARK_CYAN = CSI + f'{DARK_DIGIT};36m'
BRIGHT_CYAN = CSI + f'{BRIGHT_DIGIT};36m'

WHITE = CSI + '1m'
DARK_WHITE = CSI + f'{DARK_DIGIT};37m'
BRIGHT_WHITE = CSI + f'{BRIGHT_DIGIT};37m'

RESET_COLOR = DEFAULT_COLOR
MUTTED_COLOR = BRIGHT_BLACK

IMPLEMENTED_PROMP_COLORS = [
    DEFAULT_COLOR,
    DARK_BLACK,
    BRIGHT_BLACK,
    DARK_RED,
    BRIGHT_RED,
    DARK_GREEN,
    BRIGHT_GREEN,
    DARK_YELLOW,
    BRIGHT_YELLOW,
    DARK_BLUE,
    BRIGHT_BLUE,
    DARK_MAGENTA,
    BRIGHT_MAGENTA,
    DARK_CYAN,
    BRIGHT_CYAN,
    WHITE,
    DARK_WHITE,
    BRIGHT_WHITE,
    RESET_COLOR,
    MUTTED_COLOR
]

OPEN_COLLECTION = 'OPEN_COLLECTION'
CLOSE_COLLECTION = 'CLOSE_COLLECTION'
WITH_COLOR = True
WITHOUT_COLOR = False

NATIVE_PROMPT_COLOR = {
    TYPE_MODULE : BRIGHT_BLUE,
    TYPE_CLASS : DARK_BLUE,
    TYPE_METHOD : BRIGHT_BLUE,
    TYPE_BUILTIN_FUNCTION_OR_METHOD : BRIGHT_BLUE,
    TYPE_FUNCTION : BRIGHT_BLUE,
    TYPE_BOOLEAN : BRIGHT_WHITE,
    TYPE_STRING : DARK_YELLOW,
    TYPE_INTEGER : DARK_GREEN, ###- BRIGHT_BLACK, ###-
    TYPE_FLOAT : DARK_GREEN ###- BRIGHT_BLACK ###-
}
NONE_PROMP_COLOR = DARK_MAGENTA
QUOTE_PROMPT_COLOR = DARK_MAGENTA
COLLECTION_PROMPT_COLOR = BRIGHT_BLACK
COMA_PROMPT_COLOR = BRIGHT_BLACK
COLON_PROMPT_COLOR = BRIGHT_BLACK

COLLECTION_TYPE = {
    TYPE_DICT : {
        OPEN_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{OPEN_DICTIONARY}{RESET_COLOR}',
            WITHOUT_COLOR : OPEN_DICTIONARY
        },
        CLOSE_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{CLOSE_DICTIONARY}{RESET_COLOR}',
            WITHOUT_COLOR : CLOSE_DICTIONARY
        }
    },
    TYPE_SET : {
        OPEN_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{OPEN_DICTIONARY}{RESET_COLOR}',
            WITHOUT_COLOR : OPEN_DICTIONARY
        },
        CLOSE_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{CLOSE_DICTIONARY}{RESET_COLOR}',
            WITHOUT_COLOR : CLOSE_DICTIONARY
        }
    },
    TYPE_LIST : {
        OPEN_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{OPEN_LIST}{RESET_COLOR}',
            WITHOUT_COLOR : OPEN_LIST
        },
        CLOSE_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{CLOSE_LIST}{RESET_COLOR}',
            WITHOUT_COLOR : CLOSE_LIST
        }
    },
    TYPE_TUPLE : {
        OPEN_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{OPEN_TUPLE}{RESET_COLOR}',
            WITHOUT_COLOR : OPEN_TUPLE
        },
        CLOSE_COLLECTION : {
            WITH_COLOR : f'{COLLECTION_PROMPT_COLOR}{CLOSE_TUPLE}{RESET_COLOR}',
            WITHOUT_COLOR : CLOSE_TUPLE
        }
    }
}


SYMBOLS = {
    "@",
    "#",
    "$",
    "%",
    "&",
    "=",
    "|",
    "_",
    "/",
    "*",
    "-",
    "+"
}
PUNCTUATION = {
    ",",
    ".",
    "?",
    "}",
    "]",
    "[",
    "{",
    "!",
    "(",
    ")"
}
NUMBERS = {
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9"
}
A = {
    "À",
    "Á",
    "Ã",
    "Â",
    "à",
    "á",
    "ã",
    "â",
    "Ä",
    "ä"
}
E = {
    "È",
    "É",
    "Ê",
    "è",
    "é",
    "ê",
    "Ë",
    "ë"
}
I = {
    "Ì",
    "Í",
    "Î",
    "ì",
    "í",
    "î",
    "Ï",
    "i"
}
O = {
    "Ò",
    "Ó",
    "Õ",
    "Ô",
    "ò",
    "ó",
    "õ",
    "ô",
    "Ö",
    "ö"
}
U = {
    "Ù",
    "Ú",
    "Û",
    "ù",
    "ú",
    "û",
    "Ü",
    "ü"
}
C = {
    "Ç"
    "ç"
}
LOWER_CASE = {
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
}
UPPER_CASE = {
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z"
}
CHARACTERES = tuple(set([*A,*E,*I,*O,*U,*C,*LOWER_CASE,*UPPER_CASE,SPACE]))
UPPER_CASE_CHARACTERES = tuple(set([char for char in CHARACTERES if char == char.upper()]))



import time
from python_helper import ObjectHelper, StringHelper, SettingHelper, Constant, log, Test

# LOG_HELPER_SETTINGS = {
#     log.LOG : True,
#     log.SUCCESS : True,
#     log.SETTING : True,
#     log.DEBUG : True,
#     log.WARNING : True,
#     log.FAILURE : True,
#     log.WRAPPER : True,
#     log.ERROR : True,
#     log.TEST : False,
    # log.ENABLE_LOGS_WITH_COLORS : True,
#     SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT
# }

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.FAILURE : False,
    log.WRAPPER : False,
    log.ERROR : False,
    log.TEST : False
}

DICTIONARY_INSTANCE = {
    11: 'yolo',
    10: 2.2,
    True: False,
    'key': 'value',
    'anotherKey': {
        'key': 'value'
    },
    'aThirdKey': [
        'a',
        'b',
        {
            'c': 'd'
        },
        [
            None,
            True,
            'True',
            3.3,
            (
                2,
                '2'
            )
        ],
        {
            'key': (
                'e',
                'f',
                {
                    'g': {
                        'h': [
                            'i',
                            'j'
                        ]
                    }
                }
            )
        },
        {
            'someKey': 'someValue',
            'someOtherKey': {
                'q',
                'r',
                1,
                2,
                3,
                's'
            }
        }
    ]
}

TEST_SETTINGS = {}

@Test(
    environmentVariables={**LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def mustFilterSetting() :
    # Arrange
    expectedSingleQuoteSettingCase = 'b'
    singleQuoteSettingCase = f' " {expectedSingleQuoteSettingCase}   "   '

    expectedTripleQuoteSettingCase = 'b'
    tripleQuoteSettingCase = f''' """ {expectedTripleQuoteSettingCase}   """   '''


    expectedBothCasesSingleQuoteSettingCase = 'b'
    bothCasesSingleQuoteSettingCase = f' " {expectedBothCasesSingleQuoteSettingCase}   "   '
    bothCases = f''' """ {bothCasesSingleQuoteSettingCase}   """   '''

    someCommentsInBetween = 'b'
    someCommentsInBetweenSingle = f"  '  {someCommentsInBetween}  #  '  #   "
    someCommentsInBetweenTriple = f"""   '''   {someCommentsInBetweenSingle} #  '''     #   """

    onlyComment = Constant.HASH_TAG*4
    onlyCommentSurroundleByTriples = f'''"""{onlyComment}Something after comment token"""'''
    onlyCommentSetting = f'"{onlyCommentSurroundleByTriples}"'

    # Act
    filteredStringSingleQuoteSettingCase = StringHelper.filterString(singleQuoteSettingCase)
    filteredStringTripleQuoteSettingCase = StringHelper.filterString(tripleQuoteSettingCase)
    filteredStringBothCasesSingleQuoteSettingCase = StringHelper.filterString(bothCases)
    filteredStringSomeCommentsInBetween = StringHelper.filterString(someCommentsInBetweenTriple)
    filteredStringOnlyComment = StringHelper.filterString(onlyCommentSetting)

    # Assert
    assert expectedSingleQuoteSettingCase == filteredStringSingleQuoteSettingCase
    assert expectedTripleQuoteSettingCase == filteredStringTripleQuoteSettingCase
    assert expectedBothCasesSingleQuoteSettingCase == filteredStringBothCasesSingleQuoteSettingCase
    assert someCommentsInBetween == filteredStringSomeCommentsInBetween
    assert Constant.NOTHING == filteredStringOnlyComment

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def prettyJson_withSucces() :
    # Arrange
    simpleDictionaryInstance = {**{}, **DICTIONARY_INSTANCE}
    expected = '''{
        "11": "yolo",
        "10": 2.2,
        "true": false,
        "key": "value",
        "anotherKey": {
            "key": "value"
        },
        "aThirdKey": [
            "a",
            "b",
            {
                "c": "d"
            },
            [
                null,
                true,
                "True",
                3.3,
                [
                    2,
                    "2"
                ]
            ],
            {
                "key": [
                    "e",
                    "f",
                    {
                        "g": {
                            "h": [
                                "i",
                                "j"
                            ]
                        }
                    }
                ]
            },
            {
                "someKey": "someValue",
                "someOtherKey": [
                    1,
                    2,
                    3,
                    "q",
                    "r",
                    "s"
                ]
            }
        ]
    }'''.replace(Constant.SYSTEM_TAB,Constant.TAB_UNITS * Constant.SPACE)

    # Act
    toAssert = StringHelper.prettyJson(simpleDictionaryInstance, tabCount=1, withColors=True)

    # Assert
    toAssert = StringHelper.removeColors(toAssert)
    assert expected == toAssert

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def prettyPython_withSucces() :
    # Arrange
    simpleDictionaryInstance = {**{}, **DICTIONARY_INSTANCE}
    expected = '''{
        11: 'yolo',
        10: 2.2,
        True: False,
        'key': 'value',
        'anotherKey': {
            'key': 'value'
        },
        'aThirdKey': [
            'a',
            'b',
            {
                'c': 'd'
            },
            [
                None,
                True,
                'True',
                3.3,
                (
                    2,
                    '2'
                )
            ],
            {
                'key': (
                    'e',
                    'f',
                    {
                        'g': {
                            'h': [
                                'i',
                                'j'
                            ]
                        }
                    }
                )
            },
            {
                'someKey': 'someValue',
                'someOtherKey': {
                    1,
                    2,
                    3,
                    'q',
                    'r',
                    's'
                }
            }
        ]
    }'''.replace('\t','   ')

    # Act
    toAssert = StringHelper.prettyPython(simpleDictionaryInstance, tabCount=1, withColors=True)
    # log.debug(prettyPython_withSucces, 'does it works ' + toAssert + ' correctly?')

    # Assert
    toAssert = StringHelper.removeColors(toAssert)
    assert expected == toAssert

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def filterJson_withSucces() :
    # Arrange
    simpleDictionaryInstance = {'key':'value','anotherKey':{'key':'value'},'aThirdKey':['a','b',{'c':'d'},[None, True, 'True', 3.3, (2,'2')],{'key':('e','f',{'g':{'h':['i','j']}})},{'someKey':'someValue','someOtherKey':{'q','r',1,2,3,'s'}}]}
    expectedWithSpace = '''{        'key': 'value',        'anotherKey': {            'key': 'value'        },        'aThirdKey': [            'a',            'b',            {                'c': 'd'            },            [                None,                True,                'True',                3.3,                (                    2,                    '2'                )            ],            {                'key': (                    'e',                    'f',                    {                        'g': {                            'h': [                                'i',                                'j'                            ]                        }                    }                )            },            {                'someKey': 'someValue',                'someOtherKey': {                    1,                    2,                    3,                    'q',                    'r',                    's'                }            }        ]    }'''
    expectedWithoutSpace = '''{'key':'value','anotherKey':{'key':'value'},'aThirdKey':['a','b',{'c':'d'},[None,True,'True',3.3,(2,'2')],{'key':('e','f',{'g':{'h':['i','j']}})},{'someKey':'someValue','someOtherKey':{1,2,3,'q','r','s'}}]}'''
    filteredJson = StringHelper.prettyPython(simpleDictionaryInstance, tabCount=1, withColors=True)

    # Act
    toAssertWithSpace = StringHelper.filterJson(filteredJson)
    toAssertWithoutSpace = StringHelper.filterJson(filteredJson, extraCharacterList=[' '])

    # Assert
    toAssertWithSpace = StringHelper.removeColors(toAssertWithSpace)
    toAssertWithoutSpace = StringHelper.removeColors(toAssertWithoutSpace)
    assert expectedWithSpace == toAssertWithSpace
    assert expectedWithoutSpace == toAssertWithoutSpace

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def isLongString_withSuccess() :
    # Arrange
    tripleSinleQuotes = f'{Constant.TRIPLE_SINGLE_QUOTE}'
    tripleDoubleQuotes = f'{Constant.TRIPLE_DOUBLE_QUOTE}'
    doubleTripleSinleQuotes = f'{2*Constant.TRIPLE_SINGLE_QUOTE}'
    doubleTripleDoubleQuotes = f'{2*Constant.TRIPLE_DOUBLE_QUOTE}'
    doubleTripleSinleAndDoubleQuotes = f"{Constant.TRIPLE_SINGLE_QUOTE}{2*Constant.TRIPLE_DOUBLE_QUOTE}{Constant.TRIPLE_SINGLE_QUOTE}"
    actualLongStringWithTripleSinleQuotes = f'''{Constant.TRIPLE_SINGLE_QUOTE}
        longSring
    {Constant.TRIPLE_SINGLE_QUOTE}'''
    actualLongStringWithTripleDoubleQuotes = f'''{Constant.TRIPLE_DOUBLE_QUOTE}
        longSring
    {Constant.TRIPLE_DOUBLE_QUOTE}'''

    # Act
    toAssertTripleSinleQuotes = StringHelper.isLongString(tripleSinleQuotes)
    toAssertTripleDoubleQuotes = StringHelper.isLongString(tripleDoubleQuotes)
    toAssertDoubleTripleSinleQuotes = StringHelper.isLongString(doubleTripleSinleQuotes)
    toAssertDoubleTripleDoubleQuotes = StringHelper.isLongString(doubleTripleDoubleQuotes)
    toAssertDoubleTripleSinleAndDoubleQuotes = StringHelper.isLongString(doubleTripleSinleAndDoubleQuotes)
    toAssertActualLongStringWithTripleSinleQuotes = StringHelper.isLongString(actualLongStringWithTripleSinleQuotes)
    toAssertActualLongStringWithTripleDoubleQuotes = StringHelper.isLongString(actualLongStringWithTripleDoubleQuotes)

    # Assert
    assert toAssertTripleSinleQuotes
    assert toAssertTripleDoubleQuotes
    assert not toAssertDoubleTripleSinleQuotes
    assert not toAssertDoubleTripleDoubleQuotes
    assert not toAssertDoubleTripleSinleAndDoubleQuotes
    assert not toAssertActualLongStringWithTripleSinleQuotes
    assert not toAssertActualLongStringWithTripleDoubleQuotes

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def prettifyPerformance() :
    # arrange
    TEST_SIZE = 100
    dictionaryToPrettify = {}
    for index in range(TEST_SIZE) :
        dictionaryToPrettify[f'key_{index}'] = DICTIONARY_INSTANCE

    # act
    performanceTime = 0
    performanceTimeInit = time.time()
    toAssertPython = StringHelper.prettyPython(dictionaryToPrettify, tabCount=1, withColors=True)
    toAssertJson = StringHelper.prettyJson(dictionaryToPrettify, tabCount=1, withColors=True)
    performanceTime += time.time() - performanceTimeInit
    ###- 10000 returning f'{strInstance}{strInstance}' : 365.3402144908905 seconds
    ###- 10000 returning ''.join([strInstance, strInstance]) : 46.94538736343384 seconds

    # assert
    assert ObjectHelper.isNotNone(toAssertPython) and StringHelper.isNotBlank(toAssertPython)
    assert ObjectHelper.isNotNone(toAssertJson) and StringHelper.isNotBlank(toAssertJson)
    log.test(prettifyPerformance, f'performance time on a {len(str(dictionaryToPrettify))} dictionary size: {performanceTime} seconds', None)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def typpingTypes() :
    # arrange
    STRING = 'ACB abCdEfgh ab cd ef'
    STRING_TITLE = 'A C B Ab Cd Efgh Ab Cd Ef'

    # act
    getOnlyLettersToAssert = StringHelper.getOnlyLetters(STRING)
    toTitleToAssert = StringHelper.toTitle(STRING)

    # assert
    assert 'A' == 'a'.title()
    assert STRING.title() == StringHelper.getOnlyLetters(STRING).title(), (STRING.title(), StringHelper.getOnlyLetters(STRING).title())
    assert ObjectHelper.equals(STRING, getOnlyLettersToAssert), (STRING, getOnlyLettersToAssert)
    assert STRING_TITLE == toTitleToAssert, (STRING_TITLE, toTitleToAssert)
    assert ObjectHelper.equals(STRING_TITLE, toTitleToAssert), (STRING_TITLE, getOnlyLettersToAssert)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def toPascalCase() :
    # arrange
    STRING = 'ACB abCdEfgh ab cd ef'

    # act
    toAssert = StringHelper.toPascalCase(STRING)

    # assert
    assert ObjectHelper.equals('ACBAbCdEfghAbCdEf', toAssert), ('ACBAbCdEfghAbCdEf', toAssert)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def toCamelCase() :
    # arrange
    STRING = 'ACB abCdEfgh ab cd ef'

    # act
    toAssert = StringHelper.toCamelCase(STRING)

    # assert
    assert ObjectHelper.equals('aCBAbCdEfghAbCdEf', toAssert), ('aCBAbCdEfghAbCdEf', toAssert)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def toSnakeCase() :
    # arrange
    STRING = 'ACB abCdEfgh ab cd ef'

    # act
    toAssert = StringHelper.toSnakeCase(STRING)

    # assert
    assert ObjectHelper.equals('a_c_b_ab_cd_efgh_ab_cd_ef', toAssert), ('a_c_b_ab_cd_efgh_ab_cd_ef', toAssert)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def toKebabCase() :
    # arrange
    STRING = 'ACB abCdEfgh ab cd ef'

    # act
    toAssert = StringHelper.toKebabCase(STRING)

    # assert
    assert ObjectHelper.equals('a-c-b-ab-cd-efgh-ab-cd-ef', toAssert), ('a-b-c-ab-cd-efgh-ab-cd-ef', toAssert)

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def toTitle() :
    # arrange
    EXPECTED = 'A C B Ab Cd Efgh Ab Cd Ef'

    # act
    # assert
    assert EXPECTED == StringHelper.toTitle('aCBAbCdEfghAbCdEf')
    assert EXPECTED == StringHelper.toTitle('ACBAbCdEfghAbCdEf')
    assert EXPECTED == StringHelper.toTitle('a_c_b_ab_cd_efgh_ab_cd_ef')
    assert EXPECTED == StringHelper.toTitle('a-c-b-ab-cd-efgh-ab-cd-ef')



import random
from python_helper.api.src.service import ObjectHelper

def updateMaximum(minimum, maximum) :
    if maximum < minimum :
        return minimum
    return maximum

def sampleCollection(collection, collectionClass, length=None) :
    if ObjectHelper.isNotNone(length) :
        if ObjectHelper.isList(collection) :
            return collectionClass(random.sample(collection, length))
        return collectionClass(random.sample(list(collection), length))
    if ObjectHelper.isList(collection) :
        return random.choice(collection)
    return random.choice(list(collection))



from python_helper import SettingHelper, log, ObjectHelper, Function, Method, RandomHelper, Test

# LOG_HELPER_SETTINGS = {
#     log.LOG : True,
#     log.SUCCESS : True,
#     log.SETTING : True,
#     log.DEBUG : True,
#     log.WARNING : True,
#     log.WRAPPER : True,
#     log.FAILURE : True,
#     log.ERROR : True,
    # log.TEST : False
# }

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.WRAPPER : False,
    log.FAILURE : False,
    log.ERROR : False,
    log.TEST : False
}

TEST_SETTINGS = {
    'inspectGlobals' : False,
    'logResult' : True
}

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def Function_withSuccess() :
    # Arrange
    TEST = RandomHelper.string(minimum=10)
    @Function
    def myFunction(something) :
        return TEST, something
    @Function
    def myOtherFunction(something) :
        raise Exception(TEST)
    SOMETHING = RandomHelper.string(minimum=10)
    exception = None

    # Act
    myRestult = myFunction(SOMETHING)
    myOtherResult = None
    try :
        myOtherResult = myOtherFunction(SOMETHING)
    except Exception as ext :
        exception = ext

    # Assert
    assert (TEST, SOMETHING) == myRestult
    assert ObjectHelper.isNone(myOtherResult)
    assert ObjectHelper.isNotNone(exception)
    assert TEST == str(exception)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def Method_withSuccess() :
    # Arrange
    TEST = RandomHelper.string(minimum=10)
    class MyClass :
        @Method
        def myMethod(self,something) :
            return TEST, something
        @Method
        def myOtherMethod(self,something) :
            raise Exception(TEST)

    @Method
    def myNotMethod(self,something) :
        raise Exception(TEST)
    SOMETHING = RandomHelper.string(minimum=10)
    methodException = None
    notMethodEception = None
    myClass = MyClass()

    # Act
    myRestult = myClass.myMethod(SOMETHING)
    myOtherResult = None
    try :
        myOtherResult = myClass.myOtherMethod(SOMETHING)
    except Exception as ext :
        methodException = ext
    myNotMethodResult = None
    try :
        myNotMethodResult = myNotMethod(None,SOMETHING)
    except Exception as ext :
        notMethodEception = ext
        # print(notMethodEception)

    # Assert
    assert (TEST, SOMETHING) == myRestult
    assert ObjectHelper.isNone(myOtherResult)
    assert ObjectHelper.isNotNone(methodException)
    assert TEST == str(methodException)
    assert ObjectHelper.isNone(myNotMethodResult)
    assert ObjectHelper.isNotNone(notMethodEception)
    assert TEST == str(notMethodEception)



from python_helper import SettingHelper, StringHelper, Constant, log, EnvironmentHelper, ObjectHelper, Test, TestHelper

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : True,
    log.SETTING : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

@Test(
    environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : None,
        **LOG_HELPER_SETTINGS
    }
)
def getRaisedException_withSuccess() :
    # Arrange
    myArg = 'myArgValue'
    myKwarg = 'myKwargValue'
    expected = Exception('my exception')
    def someFunction(myArg, myKwarg='myKwargValue'):
        raise expected

    # Act
    toAssert = TestHelper.getRaisedException(someFunction, myArg, myKwarg=myKwarg)

    # Assert
    assert expected == toAssert, f'{expected} == {toAssert}: {expected == toAssert}'



from python_helper.api.src.service import SettingHelper, ReflectionHelper, ObjectHelper

def EnvironmentVariable(*outerArgs, environmentVariables=None, **outerKwargs) :
    def innerMethodWrapper(resourceInstanceMethod,*innerMethodArgs,**innerMethodKwargs) :
        def innerResourceInstanceMethod(*innerArgs,**innerKwargs) :
            methodReturn = None
            wraperException = None
            originalEnvironmentVariables, originalActiveEnvironment = SettingHelper.extractEnvironmentVariables(environmentVariables)
            try :
                methodReturn = resourceInstanceMethod(*innerArgs,**innerKwargs)
            except Exception as exception :
                wraperException = exception
            SettingHelper.recoverEnvironmentVariables(environmentVariables, originalEnvironmentVariables, originalActiveEnvironment)
            if ObjectHelper.isNotNone(wraperException) :
                raise wraperException
            return methodReturn
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceInstanceMethod)
        return innerResourceInstanceMethod
    return innerMethodWrapper



from python_helper import RandomHelper
from python_helper import SettingHelper, log, Test

# LOG_HELPER_SETTINGS = {
#     log.LOG : True,
#     log.SUCCESS : True,
#     log.SETTING : True,
#     log.DEBUG : True,
#     log.WARNING : True,
#     log.FAILURE : True,
#     log.WRAPPER : True,
#     log.ERROR : True,
    # log.TEST : False,
    # log.ENABLE_LOGS_WITH_COLORS : True,
#     SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT
# }

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.FAILURE : False,
    log.WRAPPER : False,
    log.ERROR : False,
    log.TEST : False
}

TEST_SETTINGS = {}

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def sample_withSuccess() :
    # Arrange
    SAMPLE_SIZE = 2
    listColletcion = [
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string()
    ]
    tupleColletcion = (
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string(),
        RandomHelper.string()
    )
    setColletcion = {
        RandomHelper.string(minimum=1),
        RandomHelper.string(minimum=1),
        RandomHelper.string(minimum=1),
        RandomHelper.string(minimum=1),
        RandomHelper.string(minimum=1)
    }
    dictionaryColletcion = {
        RandomHelper.string(minimum=1) : RandomHelper.string(),
        RandomHelper.string(minimum=1) : RandomHelper.string(),
        RandomHelper.string(minimum=1) : RandomHelper.string(),
        RandomHelper.string(minimum=1) : RandomHelper.string(),
        RandomHelper.string(minimum=1) : RandomHelper.string(),
    }

    # Act
    listElement = RandomHelper.sample(listColletcion)
    listSample = RandomHelper.sample(listColletcion, length=SAMPLE_SIZE)
    tupleElement = RandomHelper.sample(tupleColletcion)
    tupleSample = RandomHelper.sample(tupleColletcion, length=SAMPLE_SIZE)
    setElement = RandomHelper.sample(setColletcion)
    setSample = RandomHelper.sample(setColletcion, length=SAMPLE_SIZE)
    dictionaryElement = RandomHelper.sample(dictionaryColletcion)
    dictionarySample = RandomHelper.sample(dictionaryColletcion, length=SAMPLE_SIZE)

    # Assert
    assert listElement in listColletcion
    assert list == type(listSample)
    assert SAMPLE_SIZE == len(listSample)
    for element in listSample :
        assert element in listColletcion
    assert tupleElement in tupleColletcion
    assert tuple == type(tupleSample)
    assert SAMPLE_SIZE == len(tupleSample)
    for element in tupleSample :
        assert element in tupleColletcion
    assert setElement in setColletcion
    assert set == type(setSample)
    assert SAMPLE_SIZE == len(setSample)
    for element in setSample :
        assert element in setColletcion
    assert 1 == len(dictionaryElement.keys())
    assert list(dictionaryElement.keys())[0] in dictionaryColletcion
    assert dict == type(dictionarySample)
    assert SAMPLE_SIZE == len(dictionarySample)
    for key in dictionarySample :
        assert key in dictionaryColletcion
        assert dictionarySample[key] == dictionaryColletcion[key]

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def randomValues() :
    # Arrange
    MINIMUM_CUSTOM_RANGE = 30
    MAXIMUM_CUSTOM_RANGE = 10
    OTHER_MINIMUM_CUSTOM_RANGE = 100
    OTHER_MAXIMUM_CUSTOM_RANGE = 200
    MINIMUM_CUSTOM_FLOAT_RANGE = 400.33
    MAXIMUM_CUSTOM_FLOAT_RANGE = 500.55

    # Act
    stringValueInDefaultRange = RandomHelper.string()
    stringValueInCustomRange = RandomHelper.string(minimum=MINIMUM_CUSTOM_RANGE, maximum=MAXIMUM_CUSTOM_RANGE)
    stringValueInOtherCustomRange = RandomHelper.string(minimum=OTHER_MINIMUM_CUSTOM_RANGE, maximum=OTHER_MAXIMUM_CUSTOM_RANGE)
    integerValueInDefaultRange = RandomHelper.integer()
    integerValueInCustomRange = RandomHelper.integer(minimum=MINIMUM_CUSTOM_RANGE, maximum=MAXIMUM_CUSTOM_RANGE)
    integerValueInOtherCustomRange = RandomHelper.integer(minimum=OTHER_MINIMUM_CUSTOM_RANGE, maximum=OTHER_MAXIMUM_CUSTOM_RANGE)
    floatValueInDefaultRange = RandomHelper.float()
    floatValueInCustomRange = RandomHelper.float(minimum=MINIMUM_CUSTOM_RANGE, maximum=MAXIMUM_CUSTOM_RANGE)
    floatValueInOtherCustomRange = RandomHelper.float(minimum=OTHER_MINIMUM_CUSTOM_RANGE, maximum=OTHER_MAXIMUM_CUSTOM_RANGE)
    floatValueInOtherCustomFloatRange = RandomHelper.float(minimum=MINIMUM_CUSTOM_FLOAT_RANGE, maximum=MAXIMUM_CUSTOM_FLOAT_RANGE)

    # Assert
    assert RandomHelper.DEFAULT_MINIMUM_LENGHT <= len(stringValueInDefaultRange) <= RandomHelper.DEFAULT_MAXIMUM_LENGHT
    assert MINIMUM_CUSTOM_RANGE == len(stringValueInCustomRange)
    assert OTHER_MINIMUM_CUSTOM_RANGE <= len(stringValueInOtherCustomRange) <= OTHER_MAXIMUM_CUSTOM_RANGE
    assert RandomHelper.DEFAULT_MINIMUM_LENGHT <= integerValueInDefaultRange <= RandomHelper.DEFAULT_MAXIMUM_LENGHT
    assert MINIMUM_CUSTOM_RANGE == integerValueInCustomRange
    assert OTHER_MINIMUM_CUSTOM_RANGE <= integerValueInOtherCustomRange <= OTHER_MAXIMUM_CUSTOM_RANGE
    assert RandomHelper.DEFAULT_MINIMUM_LENGHT <= floatValueInDefaultRange <= RandomHelper.DEFAULT_MAXIMUM_LENGHT
    assert MINIMUM_CUSTOM_RANGE == floatValueInCustomRange
    assert OTHER_MINIMUM_CUSTOM_RANGE <= floatValueInOtherCustomRange < OTHER_MAXIMUM_CUSTOM_RANGE
    assert MINIMUM_CUSTOM_FLOAT_RANGE <= floatValueInOtherCustomFloatRange <= MAXIMUM_CUSTOM_FLOAT_RANGE



from python_helper.api.src.service import LogHelper, ReflectionHelper
from python_helper.api.src.domain import Constant as c

def Function(function,*args,**kwargs) :
    def wrapedFunction(*args,**kwargs) :
        try :
            functionReturn = function(*args,**kwargs)
        except Exception as exception :
            functionName = ReflectionHelper.getName(function, typeName=c.TYPE_FUNCTION)
            LogHelper.wrapper(Function,f'''Failed to execute "{functionName}(args={args}, kwargs={kwargs})" {c.TYPE_FUNCTION} call''', exception=exception, muteStackTrace=True)
            raise exception
        return functionReturn
    ReflectionHelper.overrideSignatures(wrapedFunction, function)
    return wrapedFunction

def Method(method,*args,**kwargs) :
    def wrapedMethod(*args,**kwargs) :
        try :
            methodReturn = method(*args,**kwargs)
        except Exception as exception :
            className = ReflectionHelper.getClassName(args[0].__class__, typeClass=c.TYPE_CLASS)
            methodName = ReflectionHelper.getClassName(method, typeClass=c.TYPE_METHOD)
            LogHelper.wrapper(Method,f'''Failed to execute "{className}{c.DOT}{methodName}(args={args}, kwargs={kwargs})" {c.TYPE_METHOD} call''', exception=exception, muteStackTrace=True)
            raise exception
        return methodReturn
    ReflectionHelper.overrideSignatures(wrapedMethod, method)
    return wrapedMethod



from python_helper.api.src.service import StringHelper, LogHelper, ObjectHelper, EnvironmentHelper, FileHelper
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.helper import StringHelperHelper, SettingHelperHelper

global ACTIVE_ENVIRONMENT_VALUE
ACTIVE_ENVIRONMENT_VALUE = None

ACTIVE_ENVIRONMENT = 'ACTIVE_ENVIRONMENT'
DEFAULT_ENVIRONMENT = 'default'
LOCAL_ENVIRONMENT = 'local'

def logEnvironmentSettings() :
    try :
        LogHelper.setting(logEnvironmentSettings, StringHelper.prettyJson(EnvironmentHelper.getSet()))
    except Exception as exception :
        LogHelper.failure(logEnvironmentSettings, 'Not possible do get a pretty json from EnvironmentHelper.getSet()', exception)
        LogHelper.setting(logEnvironmentSettings, EnvironmentHelper.getSet())

def updateActiveEnvironment(activeEnvironment) :
    global ACTIVE_ENVIRONMENT_VALUE
    ACTIVE_ENVIRONMENT_VALUE = DEFAULT_ENVIRONMENT if ObjectHelper.isNone(activeEnvironment) else activeEnvironment
    EnvironmentHelper.update(ACTIVE_ENVIRONMENT, ACTIVE_ENVIRONMENT_VALUE)
    return getValueAsString(ACTIVE_ENVIRONMENT_VALUE)

def softlyGetActiveEnvironment() :
    global ACTIVE_ENVIRONMENT_VALUE
    if ObjectHelper.isNone(ACTIVE_ENVIRONMENT_VALUE) :
        activeEnvironment = EnvironmentHelper.get(ACTIVE_ENVIRONMENT)
        if ObjectHelper.isNone(activeEnvironment) :
            ACTIVE_ENVIRONMENT_VALUE = updateActiveEnvironment(DEFAULT_ENVIRONMENT)
        else :
            ACTIVE_ENVIRONMENT_VALUE = activeEnvironment
    return getValueAsString(ACTIVE_ENVIRONMENT_VALUE)

def getActiveEnvironment() :
    global ACTIVE_ENVIRONMENT_VALUE
    ACTIVE_ENVIRONMENT_VALUE = updateActiveEnvironment(EnvironmentHelper.get(ACTIVE_ENVIRONMENT))
    return getValueAsString(ACTIVE_ENVIRONMENT_VALUE)

def activeEnvironmentIsDefault() :
    global ACTIVE_ENVIRONMENT_VALUE
    # return DEFAULT_ENVIRONMENT == getActiveEnvironment()
    return DEFAULT_ENVIRONMENT == ACTIVE_ENVIRONMENT_VALUE

def activeEnvironmentIsLocal() :
    global ACTIVE_ENVIRONMENT_VALUE
    # return LOCAL_ENVIRONMENT == getActiveEnvironment()
    return LOCAL_ENVIRONMENT == ACTIVE_ENVIRONMENT_VALUE

def getValueAsString(value) :
    return c.NOTHING.join([
        value,
        c.NOTHING
    ])

def extractEnvironmentVariables(environmentVariables) :
    global ACTIVE_ENVIRONMENT_VALUE
    originalActiveEnvironment = None if ObjectHelper.isNone(ACTIVE_ENVIRONMENT_VALUE) else f'{ACTIVE_ENVIRONMENT_VALUE}'
    if ObjectHelper.isNotEmpty(originalActiveEnvironment) :
        ACTIVE_ENVIRONMENT_VALUE = None
    originalEnvironmentVariables = {}
    if ObjectHelper.isDictionary(environmentVariables) :
        for key, value in environmentVariables.items() :
            originalEnvironmentVariables[key] = EnvironmentHelper.switch(key, value)
    getActiveEnvironment()
    LogHelper.loadSettings()
    return originalEnvironmentVariables, originalActiveEnvironment

def recoverEnvironmentVariables(environmentVariables, originalEnvironmentVariables, originalActiveEnvironment) :
    global ACTIVE_ENVIRONMENT_VALUE
    EnvironmentHelper.reset(environmentVariables, originalEnvironmentVariables)
    LogHelper.loadSettings()
    ACTIVE_ENVIRONMENT_VALUE = originalActiveEnvironment

def getSettingTree(
    settingFilePath,
    settingTree = None,
    fallbackSettingFilePath = None,
    fallbackSettingTree = None,
    lazyLoad = False,
    keepDepthInLongString = False,
    depthStep = c.TAB_UNITS,
    encoding = c.ENCODING
) :
    settingInjectionList = []
    fallbackSettingInjectionList = []
    if ObjectHelper.isNotNone(fallbackSettingFilePath) :
        innerFallbackSettingTree, fallbackSettingInjectionList = getSettingTree(
            fallbackSettingFilePath,
            lazyLoad = True,
            keepDepthInLongString = keepDepthInLongString,
            depthStep = depthStep,
            encoding = encoding
        )
        # print(f'innerFallbackSettingTree: {innerFallbackSettingTree}')
        # print(f'fallbackSettingInjectionList: {fallbackSettingInjectionList}')
    else :
        innerFallbackSettingTree = {}
    # print('========= in ===========')
    # print('========= in ===========')
    # print('========= in ===========')
    # print(f'innerFallbackSettingTree: {innerFallbackSettingTree}')
    # print(f'fallbackSettingInjectionList: {fallbackSettingInjectionList}')
    allSettingLines = FileHelper.getFileLines(settingFilePath, encoding=encoding)
    longStringCapturing = False
    quoteType = None
    longStringList = None
    depth = 0
    nodeRefference = 0
    nodeKey = c.NOTHING
    if settingTree is None :
        settingTree = {}
    for line, settingLine in enumerate(allSettingLines) :
        if SettingHelperHelper.lineAproved(settingLine) :
            if longStringCapturing :
                # print('lineAproved')
                if not currentDepth :
                    currentDepth = 0
                longStringList.append(depthStep*c.SPACE + settingLine if keepDepthInLongString else settingLine[depth:])
                if quoteType in str(settingLine) :
                    longStringList[-1] = c.NOTHING.join(longStringList[-1].split(quoteType))[:-1] + quoteType
                    settingValue = c.NOTHING.join(longStringList)
                    nodeKey = SettingHelperHelper.updateSettingTreeAndReturnNodeKey(settingKey,settingValue,nodeKey,settingTree, False)
                    longStringCapturing = False
                    quoteType = None
                    longStringList = None
            else :
                currentDepth = SettingHelperHelper.getDepth(settingLine)
                isSameDepth = SettingHelperHelper.nextValidLineIsAtTheSameDepthOrLess(currentDepth, line, allSettingLines)
                # print(f'isSameDepth: {isSameDepth}')
                if currentDepth == depth :
                    # print('currentDepth == depth')
                    settingKey,settingValue,nodeKey,longStringCapturing,quoteType,longStringList = SettingHelperHelper.settingTreeInnerLoop(
                        settingLine,
                        nodeKey,
                        settingTree,
                        longStringCapturing,
                        quoteType,
                        longStringList,
                        settingInjectionList,
                        lazyLoad,
                        isSameDepth
                    )
                    # print('========= out ==========')
                    # print(f'settingKey: {settingKey}, settingValue: {settingValue}')
                elif currentDepth > depth :
                    # print('currentDepth > depth')
                    currentNodeRefference = currentDepth // (currentDepth - depth)
                    if currentNodeRefference - nodeRefference == 1 :
                        settingKey,settingValue,nodeKey,longStringCapturing,quoteType,longStringList = SettingHelperHelper.settingTreeInnerLoop(
                            settingLine,
                            nodeKey,
                            settingTree,
                            longStringCapturing,
                            quoteType,
                            longStringList,
                            settingInjectionList,
                            lazyLoad,
                            isSameDepth
                        )
                        # print('========= out ==========')
                        # print(f'settingKey: {settingKey}, settingValue: {settingValue}')
                        nodeRefference = currentNodeRefference
                        depth = currentDepth
                elif currentDepth < depth :
                    # print('currentDepth < depth')
                    nodeRefference = currentDepth // depthStep ###- if line.split(':')[0] == currentDepth + c.TAB_UNITS * ' ' +
                    depth = currentDepth
                    # print()
                    # print()
                    # print()
                    # print(str(line))
                    # print(str(nodeKey))
                    # print(str(settingLine))
                    splitedNodeKey = nodeKey.split(c.DOT)[:nodeRefference]
                    splitedNodeKeyLength = len(splitedNodeKey)
                    if splitedNodeKeyLength == 0 :
                        nodeKey = c.NOTHING
                    elif splitedNodeKeyLength == 1 :
                        nodeKey = splitedNodeKey[0]
                    else :
                        nodeKey = c.DOT.join(splitedNodeKey)
                    settingKey,settingValue,nodeKey,longStringCapturing,quoteType,longStringList = SettingHelperHelper.settingTreeInnerLoop(
                        settingLine,
                        nodeKey,
                        settingTree,
                        longStringCapturing,
                        quoteType,
                        longStringList,
                        settingInjectionList,
                        lazyLoad,
                        isSameDepth
                    )
                    # print('========= out ==========')
                    # print(f'settingKey: {settingKey}, settingValue: {settingValue}')
                    # if ObjectHelper.isNotNone(settingValue) and ObjectHelper.isNotEmpty(settingValue) :
                    depth = currentDepth
    # print(f'innerFallbackSettingTree: {innerFallbackSettingTree}')
    # print(f'fallbackSettingFilePath: {fallbackSettingFilePath}')
    # print('========= out ==========')
    # print('========= out ==========')
    # print('========= out ==========')
    if lazyLoad :
        # print('Lazy load')
        # print(f'settingTree: {settingTree}')
        # print(f'settingInjectionList: {settingInjectionList}')
        return settingTree, settingInjectionList
    elif (
        ObjectHelper.isNotEmptyCollection(innerFallbackSettingTree) and
        ObjectHelper.isNotNone(fallbackSettingFilePath)
    ):
        # print('Not lazy load')
        for fallbackSettingInjection in fallbackSettingInjectionList.copy() :
            for settingInjection in settingInjectionList.copy() :
                if (
                    fallbackSettingInjection[SettingHelperHelper.SETTING_KEY] == settingInjection[SettingHelperHelper.SETTING_KEY] and
                    fallbackSettingInjection[SettingHelperHelper.SETTING_NODE_KEY] == settingInjection[SettingHelperHelper.SETTING_NODE_KEY]
                ) :
                    fallbackSettingInjectionList.remove(fallbackSettingInjection)

        # print(f'fallbackSettingInjectionList: {fallbackSettingInjectionList}')
        # print(f'settingInjectionList: {settingInjectionList}')
        settingInjectionList += fallbackSettingInjectionList
        updateSettingTree(settingTree, innerFallbackSettingTree)
        SettingHelperHelper.handleSettingInjectionList(settingInjectionList, settingTree, fallbackSettingTree=innerFallbackSettingTree)
    SettingHelperHelper.handleSettingInjectionList(settingInjectionList, settingTree, fallbackSettingTree=fallbackSettingTree)
    updateSettingTree(settingTree, fallbackSettingTree)
    return settingTree

def updateSettingTree(toUpdateSettingTree, gatheringSettingTree) :
    if ObjectHelper.isNotEmpty(gatheringSettingTree) :
        if ObjectHelper.isNone(toUpdateSettingTree) or StringHelper.isBlank(toUpdateSettingTree) :
            toUpdateSettingTree = {}
        if ObjectHelper.isCollection(gatheringSettingTree) and ObjectHelper.isDictionary(gatheringSettingTree) :
            for key, value in gatheringSettingTree.items() :
                if ObjectHelper.isNotEmpty(value) and ObjectHelper.isNotNone(value):
                    if key not in toUpdateSettingTree or ObjectHelper.isEmpty(toUpdateSettingTree[key]) :
                        # print(f'        key: {key}, value: {value}')
                        toUpdateSettingTree[key] = value
                    else :
                        updateSettingTree(toUpdateSettingTree[key], gatheringSettingTree[key])
                elif key not in toUpdateSettingTree :
                    # print(f'        toUpdateSettingTree: key: {key}, value: {value}')
                    toUpdateSettingTree[key] = value


def getSetting(nodeKey,settingTree) :
    setting = None
    try :
        setting = SettingHelperHelper.accessTree(nodeKey,settingTree)
    except Exception as exception :
        LogHelper.failure(getSetting, f'Not possible to get {nodeKey} node key. Returning "{setting}" by default', exception)
    return setting if not isinstance(setting, str) else StringHelper.filterString(setting)

def querySetting(keywordQuery,tree) :
    if StringHelper.isBlank(keywordQuery) or ObjectHelper.isNotDictionary(tree) :
        LogHelper.warning(querySetting,f'''Not possible to parse "{tree}". It's either is not a dictionary or "{keywordQuery}" keyword query is blank''')
    querySet = {}
    SettingHelperHelper.keepSearching(keywordQuery,tree,querySet)
    return querySet

def printSettings(tree, name, depth=1, withColors=False):
    withColors = LogHelper.colorsEnabled() ###- activeEnvironmentIsLocal()
    settingKeyColor = SettingHelperHelper.getSettingKeyPrompColor(withColors)
    colonColor = SettingHelperHelper.getSettingColonPrompColor(withColors)
    print(f'{c.NEW_LINE}{settingKeyColor}{c.OPEN_LIST}{name.upper()}{c.CLOSE_LIST}{colonColor}{c.SPACE}{c.COLON}')
    SettingHelperHelper.printNodeTree(
        tree,
        depth,
        settingKeyColor = settingKeyColor,
        settingValueColor = SettingHelperHelper.getSettingValuePrompColor(withColors),
        colonColor = colonColor,
        resetColor = SettingHelperHelper.getSettingResetPrompColor(withColors)
    )
    print()



from numbers import Number
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.service import ObjectHelper, StringHelper, ReflectionHelper

def getColorValue(thing, color) :
    return color if ObjectHelper.isNotNone(color) else c.NATIVE_PROMPT_COLOR.get(
        ReflectionHelper.getName(type(thing)),
        c.DARK_BLUE
    )

def newLine(strReturn, character, prettyFunction, withColors):
    if character == strReturn[-1][-len(character):] :
        return [c.NEW_LINE]
    else :
        return [*getItAsColoredString(c.COMA, prettyFunction, withColors, color=c.COMA_PROMPT_COLOR), *c.NEW_LINE]

def getValueCollection(outterValue) :
    return outterValue if not isinstance(outterValue, set) else ObjectHelper.getSortedCollection(outterValue)

def getItAsColoredString(thing, prettyFunction, withColors, replaceBy=None, color=None) :
    thingValue = str(thing) if ObjectHelper.isNone(replaceBy) else str(replaceBy)
    return [thingValue] if not withColors else [getColorValue(thing, color), thingValue, c.RESET_COLOR]

def getFilteredAndColoredQuote(keyOrValue, string, prettyFunction, withColors, color) :
    if ObjectHelper.isNativeClassInstance(keyOrValue) and not isinstance(keyOrValue, str) :
        return [c.NOTHING] if StringHelper.prettyPython == prettyFunction else getItAsColoredString(string, prettyFunction, withColors, color=color)
    return getItAsColoredString(string, prettyFunction, withColors, color=color)

def getStrReturn(
        key,
        value,
        collectionType,
        quote,
        prettyFunction,
        tabCount,
        nullValue,
        trueValue,
        falseValue,
        withColors
    ) :
    valueValue = prettyFunction(
        value,
        quote = quote,
        tabCount = tabCount,
        nullValue = nullValue,
        trueValue = trueValue,
        falseValue = falseValue,
        withColors = withColors,
        joinAtReturn = False
    )
    if c.TYPE_DICT == collectionType :
        filteredAndColoredQuote = getFilteredAndColoredQuote(key, quote, prettyFunction, withColors, c.QUOTE_PROMPT_COLOR)
        filteredAndColoredColonSpace = getItAsColoredString(c.COLON_SPACE, prettyFunction, withColors, color=c.COLON_PROMPT_COLOR)
        return [tabCount * c.TAB, *filteredAndColoredQuote, *getItAsColoredString(key if StringHelper.prettyPython==prettyFunction else StringHelper.filterString(prettyFunction(key, quote=quote, tabCount=tabCount, nullValue=nullValue, trueValue=trueValue, falseValue=falseValue)), prettyFunction, withColors), *filteredAndColoredQuote, *filteredAndColoredColonSpace, *valueValue]
    else :
        return [tabCount * c.TAB, *valueValue]

def prettyInstance(
        outterValue,
        quote,
        prettyFunction,
        tabCount,
        nullValue,
        trueValue,
        falseValue,
        withColors=False,
        joinAtReturn=False
    ) :
    strReturn = []
    if (isinstance(outterValue, int) or isinstance(outterValue, float)) and not isinstance(outterValue, bool) :
        strReturn += getItAsColoredString(outterValue, prettyFunction, withColors)
    elif isinstance(outterValue, bool) :
        if True == outterValue:
            strReturn += getItAsColoredString(outterValue, prettyFunction, withColors, replaceBy=trueValue)
        elif False == outterValue:
            strReturn += getItAsColoredString(outterValue, prettyFunction, withColors, replaceBy=falseValue)
    elif ObjectHelper.isNone(outterValue) :
        strReturn += getItAsColoredString(outterValue, prettyFunction, withColors, replaceBy=nullValue, color=c.NONE_PROMP_COLOR)
    else :
        filteredAndColoredQuote = getFilteredAndColoredQuote(outterValue, quote, prettyFunction, withColors, c.QUOTE_PROMPT_COLOR)
        strReturn += [*filteredAndColoredQuote, *getItAsColoredString(outterValue, prettyFunction, withColors), *filteredAndColoredQuote]
    return returnStr(strReturn, joinAtReturn)

def getCollectionCharacter(collectionType, openOrClose, withColors) :
    return c.COLLECTION_TYPE.get(
        collectionType,
        c.COLLECTION_TYPE.get(c.TYPE_LIST)
    ).get(openOrClose).get(
        withColors,
        c.COLLECTION_TYPE.get(
            collectionType,
            c.COLLECTION_TYPE.get(c.TYPE_LIST)
        ).get(openOrClose).get(c.WITHOUT_COLOR)
    )

def prettyCollection(
        outterValue,
        collectionType,
        quote,
        prettyFunction,
        tabCount,
        nullValue,
        trueValue,
        falseValue,
        withColors=False,
        joinAtReturn=False
    ) :
    openCollection = getCollectionCharacter(collectionType, c.OPEN_COLLECTION, withColors)
    closeCollection = getCollectionCharacter(collectionType, c.CLOSE_COLLECTION, withColors)
    strReturn = []
    if len(outterValue) == 0 :
        strReturn += [openCollection, closeCollection]
    else :
        strReturn += [openCollection]
        tabCount += 1
        if ObjectHelper.isDictionary(outterValue) :
            for key,value in outterValue.items() :
                strReturn += newLine(strReturn, openCollection, prettyFunction, withColors)
                strReturn += getStrReturn(key, value, collectionType, quote, prettyFunction, tabCount, nullValue, trueValue, falseValue, withColors)
        else :
            for value in outterValue :
                strReturn += newLine(strReturn, openCollection, prettyFunction, withColors)
                strReturn += getStrReturn(None, value, collectionType, quote, prettyFunction, tabCount, nullValue, trueValue, falseValue, withColors)
        strReturn += [c.NEW_LINE]
        tabCount -= 1
        strReturn += [tabCount * c.TAB, closeCollection]
    return returnStr(strReturn, joinAtReturn)

def returnStr(strReturn, joinAtReturn) :
    return strReturn if not joinAtReturn else StringHelper.join(strReturn)

def isNotOneLineLongString(quoteType, thing) :
    if isinstance(thing, str) and isinstance(quoteType, str) and quoteType in [c.TRIPLE_SINGLE_QUOTE, c.TRIPLE_DOUBLE_QUOTE] :
        return not 0 == thing.count(quoteType) % 2
    else :
        return True



from python_helper import log
from python_helper import ObjectHelper
from python_helper import SettingHelper
from python_helper import StringHelper
from python_helper import EnvironmentHelper
from python_helper import ReflectionHelper
from python_helper import RandomHelper
from python_helper import Constant
from python_helper import EnvironmentVariable
from python_helper import Test
from python_helper import Method
from python_helper import Function
from python_helper import ObjectHelperHelper
from python_helper import SettingHelperHelper
from python_helper import SettingHelperHelper
from python_helper import LogHelperHelper
from python_helper import RandomHelperHelper
from python_helper import FileHelper

@Test(inspectGlobals=False)
def allImportedSuccesfuly() :
    ...



from python_helper import EnvironmentHelper, SettingHelper, ObjectHelper, log, Test

# LOG_HELPER_SETTINGS = {
#     log.LOG : True,
#     log.SUCCESS : True,
#     log.SETTING : True,
#     log.DEBUG : True,
#     log.WARNING : True,
#     log.WRAPPER : True,
#     log.FAILURE : True,
#     log.ERROR : True,
    # log.TEST : False
# }

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.WRAPPER : False,
    log.FAILURE : False,
    log.ERROR : False,
    log.TEST : False
}

TEST_SETTINGS = {}

@Test(
    environmentVariables={
        **LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def osIdentifierTest() :
    # arrange

    # act
    isWindows = EnvironmentHelper.isWindows()
    isLinux = EnvironmentHelper.isLinux()

    # assert
    assert ObjectHelper.isNotNone(isWindows)
    assert ObjectHelper.isNotNone(isLinux)
    assert ObjectHelper.isNativeClassInstance(isWindows)
    assert ObjectHelper.isNativeClassInstance(isLinux)
    assert bool == type(isWindows)
    assert bool == type(isLinux)
    assert isLinux or isWindows



from python_helper.api.src.service import StringHelper, LogHelper, ObjectHelper, EnvironmentHelper, SettingHelper
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.helper import StringHelperHelper

OPEN_SETTING_INJECTION = '${'
CLOSE_SETTING_INJECTION = '}'

SETTING_KEY = 'SETTING_KEY'
SETTING_VALUE = 'SETTING_VALUE'
SETTING_NODE_KEY = 'SETTING_NODE_KEY'

def getFilteredSetting(settingKey, settingValue, nodeKey, settingTree) :
    # print(f'getFilteredSetting: settingKey: {settingKey}, settingValue: "{settingValue}", type: {type(settingValue)}')
    if StringHelper.isNotBlank(settingValue) :
        settingEvaluationList = settingValue.split(c.COLON)
        if len(settingEvaluationList) > 1 :
            defaultSettingValue = c.COLON.join(settingEvaluationList[1:]).strip()
            # print(f'    defaultSettingValue: {defaultSettingValue}')
        else :
            # defaultSettingValue = c.NONE
            defaultSettingValue = settingValue
        if isSettingInjection(defaultSettingValue) :
            # print(f'        {settingKey} -> is setting injection: settingValue: {settingValue}, type: {type(settingValue)}')
            return getSettingInjectionValue(settingKey, defaultSettingValue, nodeKey, settingTree)
        # print(f'        {settingKey} -> not setting injection: settingValue: {settingValue}, type: {type(settingValue)}')
        return getValue(StringHelper.filterString(defaultSettingValue), ignoreCollections=True)
    # print(f'        {settingKey} -> straight: settingValue: {settingValue}, type: {type(settingValue)}')
    return settingValue

def lineAproved(settingLine) :
    approved = True
    if c.NEW_LINE == settingLine  :
        approved = False
    if c.HASH_TAG in settingLine :
        filteredSettingLine = StringHelper.filterString(settingLine)
        if filteredSettingLine is None or c.NOTHING == filteredSettingLine or c.NEW_LINE == filteredSettingLine :
            approved = False
    return approved

def nextValidLineIsAtTheSameDepthOrLess(currentDepth, line, allSettingLines) :
    ###- currentDepth == depth and
    if line + 1 == len(allSettingLines) :
        return True
    for settingLine in allSettingLines[line+1:] :
        if lineAproved(settingLine) :
            # print(f'currentDepth: {currentDepth}, line: {line}, isValid: {currentDepth == getDepth(settingLine)}, settingLine: {settingLine}')
            return currentDepth >= getDepth(settingLine)
    # print(f'currentDepth: {currentDepth}, line: {line}, isValid: {False}')
    return False

def updateSettingTreeAndReturnNodeKey(settingKey, settingValue, nodeKey, settingTree, isSameDepth) :
    # print(f'nodeKey: {nodeKey}, settingKey: {settingKey}')
    updateSettingValue(settingKey, settingValue, nodeKey, settingTree, isSameDepth=isSameDepth)
    if not isSettingValue(settingValue) :
        if c.NOTHING == nodeKey :
            nodeKey += f'{settingKey}'
        else :
            nodeKey += f'{c.DOT}{settingKey}'
    return nodeKey

def accessTree(nodeKey,tree) :
    if ObjectHelper.isNotNone(tree) :
        strippedNodeKey = nodeKey.strip()
        if ObjectHelper.isEmpty(nodeKey) :
            returnTree = None
            try :
                returnTree = StringHelper.filterString(tree)
            except Exception as exception :
                LogHelper.failure(accessTree, f'Failed to get filtered string from {tree} tree. Returning it the way it is by default', exception)
                returnTree = tree
            return returnTree
        elif isinstance(nodeKey,str) :
            nodeKeyList = nodeKey.split(c.DOT)
            if len(nodeKeyList) == 1 :
                 nextNodeKey = c.NOTHING
            else :
                nextNodeKey = c.DOT.join(nodeKeyList[1:])
            return accessTree(nextNodeKey,tree.get(nodeKeyList[0]))

def safelyAccessTree(nodeKey, settingTree) :
    setting = None
    try :
        setting = accessTree(nodeKey,settingTree)
    except Exception as exception :
        LogHelper.log(safelyAccessTree, f'Not possible to safely access "{nodeKey}" node key while looping through setting tree. Returning "{setting}" by default', exception=exception)
    return setting

def updateSettingValue(
    settingKey,
    settingValue,
    nodeKey,
    settingTree,
    isSameDepth = False,
    isSettingInjection = False
) :
    # print(f'settingKey: {settingKey}, settingValue: {settingValue}, nodeKey: {nodeKey}, settingTree: {settingTree}, isSameDepth: {isSameDepth}')
    if StringHelper.isNotBlank(settingKey) and ObjectHelper.isNotNone(settingKey) :
        # print(f'accessTree({nodeKey},{settingTree})[{settingKey}]: {accessTree(nodeKey,settingTree)}[{settingKey}]')
        # print(f'type(accessTree(nodeKey,settingTree)): {type(accessTree(nodeKey,settingTree))}')

        accessTree(nodeKey,settingTree)[settingKey] = getSettingValueOrNewNode(
            nodeKey,
            settingKey,
            settingValue,
            isSameDepth=isSameDepth,
            isSettingInjection=isSettingInjection
        )
    elif StringHelper.isNotBlank(nodeKey) and ObjectHelper.isNotNone(nodeKey) :
        splittedNodeKey = nodeKey.split(c.DOT)
        if 1 < len(splittedNodeKey) :
            accessSettingKey = splittedNodeKey[-1]
            accessNodeKey = c.DOT.join(splittedNodeKey[:-1])
        else :
            accessSettingKey = splittedNodeKey[-1]
            accessNodeKey = None
        accessTree(accessNodeKey,settingTree)[accessSettingKey] = getSettingValueOrNewNode(
            nodeKey,
            settingKey,
            settingValue,
            isSameDepth=isSameDepth,
            isSettingInjection=isSettingInjection
        )
    else :
        errorMessage = 'Node key and setting key cannot be None at the same time'
        exception = Exception(errorMessage)
        LogHelper.error(updateSettingValue, f'Error parsing settingKey: "{settingKey}", settingValue: {settingValue}, nodeKey: {nodeKey}, settingTree: {StringHelper.prettyPython(settingTree)}', errorMessage)
        raise exception

def getDepth(settingLine) :
    depthNotFount = True
    depth = 0
    while not settingLine[depth] == c.NEW_LINE and depthNotFount:
        if settingLine[depth] == c.SPACE :
            depth += 1
        else :
            depthNotFount = False
    return depth

def settingTreeInnerLoop(
    settingLine,
    nodeKey,
    settingTree,
    longStringCapturing,
    quoteType,
    longStringList,
    settingInjectionList,
    lazyLoad,
    isSameDepth
) :
    # print(f'settingTreeInnerLoop: settingLine: {settingLine}')
    # print(f'settingTreeInnerLoop: nodeKey: {nodeKey}')
    # print(f'settingTreeInnerLoop: settingTree: {settingTree}')
    # print(f'    isSameDepth: {isSameDepth}')
    settingKey, settingValue = getAttributeKeyValue(settingLine)
    # print(f'    settingKey: {settingKey}, settingValue: {settingValue}')
    # print(f'    value:{settingLine}')

    if containsSettingInjection(settingValue) :
        try :
            settingKey, settingValue, nodeKey, longStringCapturing, quoteType, longStringList = handleSettingInjection(
                settingKey,
                settingValue,
                nodeKey,
                settingTree,
                longStringCapturing,
                quoteType,
                longStringList,
                settingInjectionList,
                lazyLoad,
                isSameDepth
            )
        except Exception as exception :
            LogHelper.log(settingTreeInnerLoop, f'Not possible to handle association of "{nodeKey}{c.DOT}{settingKey}" setting key to "{settingValue}" setting value', exception=exception)
            settingInjectionList.append({
                SETTING_KEY : settingKey,
                SETTING_VALUE : settingValue,
                SETTING_NODE_KEY : nodeKey
            })
        return settingKey, settingValue, nodeKey, longStringCapturing, quoteType, longStringList
    else :
        # print(f'    nodeKey: {nodeKey}')
        settingKey, filteredSettingValue, nodeKey, longStringCapturing, quoteType, longStringList = handleLongStringOrSetting(
            settingKey,
            settingValue,
            nodeKey,
            settingTree,
            longStringCapturing,
            quoteType,
            longStringList,
            settingInjectionList,
            isSameDepth
        )
        # print(f'----------> nodeKey: {nodeKey}')
        return settingKey, filteredSettingValue, nodeKey, longStringCapturing, quoteType, longStringList

def handleSettingInjection(
    settingKey,
    settingValue,
    nodeKey,
    settingTree,
    longStringCapturing,
    quoteType,
    longStringList,
    settingInjectionList,
    lazyLoad,
    isSameDepth
) :
    # print(f'handleSettingInjection: settingKey: {settingKey}, settingValue: {settingValue}')
    if isSettingInjection(settingValue) :
        if lazyLoad :
            settingInjectionList.append({
                SETTING_KEY : settingKey,
                SETTING_VALUE : settingValue,
                SETTING_NODE_KEY : nodeKey
            })
            return settingKey, settingValue, nodeKey, longStringCapturing, quoteType, longStringList
        else :
            settingValue = getSettingInjectionValue(settingKey, settingValue, nodeKey, settingTree)
    if containsSettingInjection(settingValue) :
        settingInjectionList.append({
            SETTING_KEY : settingKey,
            SETTING_VALUE : settingValue,
            SETTING_NODE_KEY : nodeKey
        })
        return settingKey, settingValue, nodeKey, longStringCapturing, quoteType, longStringList
    return handleLongStringOrSetting(
        settingKey,
        settingValue,
        nodeKey,
        settingTree,
        longStringCapturing,
        quoteType,
        longStringList,
        settingInjectionList,
        isSameDepth
    )

def handleSettingInjectionList(settingInjectionList, settingTree, fallbackSettingTree=None) :
    # print(f'handleSettingInjectionList: settingTree: {settingTree}')
    if ObjectHelper.isNotEmptyCollection(settingInjectionList) and ObjectHelper.isNotNone(settingTree) :
        try :
            done, stuck = handleSettingInjectionListLoop(
                False,
                False,
                settingTree,
                settingInjectionList,
                fallbackSettingTree,
                getSettingInjectionValueIgnoringFallbackSettingTree,
                verifyer = verifyDefaultSettingInjectionListHandler
            )
        except Exception as exception :
            LogHelper.error(handleSettingInjectionList,'Not possible to load setting injections properly',exception)
            raise exception

def lazillyHandleSettingInjectionList(settingInjectionList, settingTree, fallbackSettingTree=None) :
    if ObjectHelper.isNotEmptyCollection(settingInjectionList) and ObjectHelper.isNotNone(settingTree) :
        try :
            done, stuck = handleSettingInjectionListLoop(
                False,
                False,
                settingTree,
                settingInjectionList,
                fallbackSettingTree,
                getSettingInjectionValueIgnoringFallbackSettingTree,
                verifyer = None
            )
        except Exception as exception :
            LogHelper.error(handleSettingInjectionList,'Not possible to lazilly load setting injections properly',exception)
            raise exception

################################################################################
################################################################################
################################################################################

def handleSettingInjectionListLoop(
    done,
    stuck,
    settingTree,
    settingInjectionList,
    fallbackSettingTree,
    handler,
    verifyer = None
) :
    # print(f'handleSettingInjectionListLoop: handler: {handler}')
    # print(f'handleSettingInjectionListLoop: settingTree: {settingTree}')
    while not done and not stuck :
        stuck = True
        isSettingInjectionCount = 0
        containsSettingInjectionCount = 0
        for settingInjection in settingInjectionList.copy() :
            done, stuck, isSettingInjectionCount, containsSettingInjectionCount = handleSettingInjectionListByCallingHandler(
                done,
                stuck,
                isSettingInjectionCount,
                containsSettingInjectionCount,
                settingInjection,
                settingTree,
                settingInjectionList,
                fallbackSettingTree,
                handler
            )
        if ObjectHelper.isNotNone(verifyer) :
            verifyer(done, stuck, isSettingInjectionCount, containsSettingInjectionCount, settingTree, settingInjectionList, fallbackSettingTree)
    # print(f'handleSettingInjectionListLoop: returning settingTree: {settingTree}')
    return done, stuck

def verifyDefaultSettingInjectionListHandler(
    done,
    stuck,
    isSettingInjectionCount,
    containsSettingInjectionCount,
    settingTree,
    settingInjectionList,
    fallbackSettingTree
) :
    if 0 == len(settingInjectionList) :
        done = True
    elif stuck :
        stuck = False
        done, stuck = handleSettingInjectionListLoop(
            done,
            stuck,
            settingTree,
            settingInjectionList,
            fallbackSettingTree,
            safelyGetSettingInjectionValue,
            verifyer = None
        )
        if 0 == len(settingInjectionList) :
            done = True
        elif stuck :
            LogHelper.setting(verifyDefaultSettingInjectionListHandler, f'Parsed settings: {StringHelper.prettyPython(settingTree)}')
            notParsedSettingInjectionDictionary = {}
            for settingInjection in settingInjectionList :
                notParsedSettingInjectionDictionary[f'{settingInjection[SETTING_NODE_KEY]}{c.DOT}{settingInjection[SETTING_KEY]}'] = (settingInjection)
            if 0 == isSettingInjectionCount and 0 == containsSettingInjectionCount :
                raise Exception(f'Circular reference detected in following setting injections: {StringHelper.prettyPython(notParsedSettingInjectionDictionary)}{c.NEW_LINE}FallbackSettingTree: {StringHelper.prettyPython(fallbackSettingTree)}{c.NEW_LINE}SettingTree: {StringHelper.prettyPython(settingTree)}')
            elif not 0 == len(settingInjectionList) :
                raise Exception(f'Not possible to parse the following setting injections: {StringHelper.prettyPython(notParsedSettingInjectionDictionary)}{c.NEW_LINE}FallbackSettingTree: {StringHelper.prettyPython(fallbackSettingTree)}{c.NEW_LINE}SettingTree: {StringHelper.prettyPython(settingTree)}')

def handleSettingInjectionListByCallingHandler(
    done,
    stuck,
    isSettingInjectionCount,
    containsSettingInjectionCount,
    settingInjection,
    settingTree,
    settingInjectionList,
    fallbackSettingTree,
    handler
) :
    # print()
    # print()
    # print(f'handleSettingInjectionListByCallingHandler: handler: {handler}')
    # print(f'handleSettingInjectionListByCallingHandler: settingTree: {settingTree}')
    try :
        if isSettingInjection(settingInjection[SETTING_VALUE]) :
            settingInjection[SETTING_VALUE] = handler(
                settingInjection[SETTING_KEY],
                settingInjection[SETTING_VALUE],
                settingInjection[SETTING_NODE_KEY],
                settingTree,
                fallbackSettingTree = fallbackSettingTree
            )
            settingInjectionArgs = list(settingInjection.values()) + [settingTree]
            updateSettingValue(*settingInjectionArgs, isSettingInjection=True)
            if not containsSettingInjection(settingInjection[SETTING_VALUE]) :
                settingInjectionList.remove(settingInjection)
                isSettingInjectionCount += 1
            stuck = False
        elif containsSettingInjection(settingInjection[SETTING_VALUE]) :
            settingInjectionListFromSettingValue = getSettingInjectionListFromSettingValue(settingInjection[SETTING_VALUE])
            newSettingInjection = settingInjection[SETTING_VALUE]
            for settingValue in settingInjectionListFromSettingValue :
                newSettingValue = handler(
                    settingInjection[SETTING_KEY],
                    settingValue,
                    settingInjection[SETTING_NODE_KEY],
                    settingTree,
                    fallbackSettingTree = fallbackSettingTree
                )
                if ObjectHelper.isList(newSettingInjection) :
                    for index, element in enumerate(newSettingInjection):
                        if settingValue == element :
                            newSettingInjection[index] = newSettingValue
                else :
                    newSettingInjection = newSettingInjection.replace(settingValue,str(newSettingValue))
            settingInjection[SETTING_VALUE] = newSettingInjection
            settingInjectionArgs = list(settingInjection.values()) + [settingTree]
            updateSettingValue(*settingInjectionArgs, isSettingInjection=True)
            if not containsSettingInjection(settingInjection[SETTING_VALUE]) :
                settingInjectionList.remove(settingInjection)
                containsSettingInjectionCount += 1
            stuck = False
    except Exception as exception :
        LogHelper.log(handleSettingInjectionListByCallingHandler, f'Ignored exception while handling {StringHelper.prettyPython(settingInjection)} setting injection list', exception=exception)
    # print(f'handleSettingInjectionListByCallingHandler: returning settingTree: {settingTree}')
    return done, stuck, isSettingInjectionCount, containsSettingInjectionCount

################################################################################
################################################################################
################################################################################

def appendSettingKey(nodeKey,settingKey):
    if StringHelper.isNotBlank(nodeKey) :
        if StringHelper.isNotBlank(settingKey) :
            return f'{nodeKey}{c.DOT}{settingKey}'
        else :
            return nodeKey
    elif StringHelper.isNotBlank(settingKey) :
        return settingKey
    else :
        return None

def safelyGetSettingInjectionValue(settingKey, settingValue, nodeKey, settingTree, fallbackSettingTree=None) :
    exception = None
    newSettingValue = None
    try :
        newSettingValue = getSettingInjectionValue(settingKey, settingValue, nodeKey, settingTree)
    except Exception as e :
        exception = e
        LogHelper.log(safelyGetSettingInjectionValue, f'Not possible to load "{settingKey}" setting key from setting tree. Now trying to load it from fallback setting tree', exception=exception)
    if ObjectHelper.isNone(newSettingValue) and ObjectHelper.isNotNone(fallbackSettingTree) :
        return getSettingInjectionValue(settingKey, settingValue, nodeKey, fallbackSettingTree)
    if ObjectHelper.isNotNone(exception) :
        raise exception
    return newSettingValue

def handleLongStringOrSetting(
    settingKey,
    settingValue,
    nodeKey,
    settingTree,
    longStringCapturing,
    quoteType,
    longStringList,
    settingInjectionList,
    isSameDepth
) :
    # print(f'handleLongStringOrSetting: settingValue: {settingValue}')
    if StringHelper.isLongString(settingValue) :
        longStringCapturing = True
        splitedSettingValueAsString = settingValue.split(c.TRIPLE_SINGLE_QUOTE)
        if c.TRIPLE_SINGLE_QUOTE in settingValue and splitedSettingValueAsString and c.TRIPLE_DOUBLE_QUOTE not in splitedSettingValueAsString[0] :
            quoteType = c.TRIPLE_SINGLE_QUOTE
        else :
            quoteType = c.TRIPLE_DOUBLE_QUOTE
        longStringList = [settingValue + c.NEW_LINE]
    else :
        nodeKey = updateSettingTreeAndReturnNodeKey(settingKey, settingValue, nodeKey, settingTree, isSameDepth)
    filteredSettingValue = getFilteredSetting(settingKey, settingValue, nodeKey, settingTree)
    # print(f'handleLongStringOrSetting: settingKey: {settingKey}, settingValue: {settingValue}, type: {type(settingValue)} filteredSettingValue: {filteredSettingValue}, type: {type(settingValue)}')
    # print(longStringCapturing, quoteType, longStringList)
    return settingKey, filteredSettingValue, nodeKey, longStringCapturing, quoteType, longStringList

def getAttributeKeyValue(settingLine) :
    settingKey = getAttributeKey(settingLine)
    settingValue = getAttibuteValue(settingLine)
    return settingKey,settingValue

def getAttributeKey(settingLine) :
    possibleKey = StringHelper.filterString(settingLine)
    return possibleKey.split(c.COLON)[0].strip()

def getAttibuteValue(settingLine) :
    possibleValue = StringHelper.filterString(settingLine)
    return getValue(c.COLON.join(possibleValue.split(c.COLON)[1:]))

def isSettingKey(possibleSettingKey) :
    return ObjectHelper.isNotNone(possibleSettingKey) and isinstance(possibleSettingKey, str) and (
        not isSettingInjection(possibleSettingKey) and
        possibleSettingKey == possibleSettingKey.lower() and
        not c.COLON in possibleSettingKey
    )

def isSettingValue(settingValue) :
    return ObjectHelper.isNotEmpty(settingValue) or ObjectHelper.isCollection(settingValue)

def getSettingValueOrNewNode(nodeKey, settingKey, settingValue, isSameDepth=False, isSettingInjection=False) :
    isActuallyASettingValue = isSettingValue(settingValue)
    # print(f'settingValue: {settingValue}, isActuallyASettingValue: {isActuallyASettingValue}, isSameDepth: {isSameDepth}')
    if not isActuallyASettingValue and isSameDepth :
        raise Exception(f'The key: "{nodeKey}{c.DOT}{settingKey}" has the following invalid value: "{settingValue}"')
    return settingValue if isActuallyASettingValue else None if isSettingInjection or isSameDepth else dict()

def getValue(value, ignoreCollections=False) :
    filteredValue = StringHelper.filterString(value)
    if isSettingValue(filteredValue) :
        if not ignoreCollections:
            if StringHelper.isNotBlank(filteredValue) :
                if c.OPEN_LIST == filteredValue[0] :
                    return getList(filteredValue)
                elif c.OPEN_TUPLE == filteredValue[0] :
                    return getTuple(filteredValue)
                elif c.OPEN_DICTIONARY == filteredValue[0] :
                    return getDictionary(filteredValue)
                elif c.OPEN_SET == filteredValue[0] :
                    return getSet(filteredValue)
        parsedValue = None
        try :
            parsedValue = int(filteredValue)
        except :
            try :
                parsedValue = float(filteredValue)
            except :
                try :
                    parsedValue = filteredValue
                    if not filteredValue is None :
                        if filteredValue == c.TRUE :
                            parsedValue = True
                        elif filteredValue == c.FALSE :
                            parsedValue = False
                except:
                    parsedValue = filteredValue
        return parsedValue
    return filteredValue

def getList(value) :
    roughtValueList = value[1:-1].split(c.COMA)
    valueList = list()
    for value in roughtValueList :
        gottenValue = getValue(value)
        if ObjectHelper.isNotEmpty(gottenValue) :
            valueList.append(gottenValue)
    return valueList

def getTuple(value) :
    return tuple(getList(value))

def getSet(value) :
    roughtValueList = value[1:-1].split(c.COMA)
    valueSet = set()
    for value in roughtValueList :
        gottenValue = getValue(value)
        if ObjectHelper.isNotNone(gottenValue) :
            valueSet.add(gottenValue)
    return valueSet

def getDictionary(value) :
    splitedValue = value[1:-1].split(c.COLON)
    keyList = []
    for index in range(len(splitedValue) -1) :
        keyList.append(StringHelper.filterString(splitedValue[index].split(c.COMA)[-1].strip()))
    valueList = []
    valueListSize = len(splitedValue) -1
    for index in range(valueListSize) :
        if index == valueListSize -1 :
            correctValue = splitedValue[index+1].strip()
        else :
            correctValue = c.COMA.join(splitedValue[index+1].split(c.COMA)[:-1]).strip()
        valueList.append(getValue(correctValue))
    resultantDictionary = dict()
    for index in range(len(keyList)) :
        resultantDictionary[keyList[index]] = valueList[index]
    return resultantDictionary

def getSettingInjectionListFromSettingValue(settingValue) :
    if ObjectHelper.isList(settingValue) :
        allElements = []
        for element in settingValue :
            elements = getSettingInjectionListFromSettingValue(element)
            if ObjectHelper.isList(elements) :
                allElements += elements
        return allElements
    elif ObjectHelper.isNotNone(settingValue) and StringHelper.isNotBlank(settingValue) :
        splitedSettingValue = settingValue.split(OPEN_SETTING_INJECTION)
        settingValueList = []
        completeSettingValue = c.NOTHING
        for segment in splitedSettingValue if settingValue.startswith(OPEN_SETTING_INJECTION) else splitedSettingValue[1:] :
            if ObjectHelper.isNotNone(segment) and StringHelper.isNotBlank(segment) :
                if ObjectHelper.isNotNone(segment.count(c.OPEN_DICTIONARY)) and not segment.count(c.OPEN_DICTIONARY) == segment.count(c.CLOSE_DICTIONARY) and 0 < segment.count(c.OPEN_DICTIONARY) :
                    completeSettingValue += segment
                else :
                    splitedSegment = segment.split(CLOSE_SETTING_INJECTION)
                    completeSettingValue += splitedSegment[0]
                    if ObjectHelper.isNotNone(completeSettingValue) and StringHelper.isNotBlank(completeSettingValue) :
                        settingValueList.append(f'{OPEN_SETTING_INJECTION}{completeSettingValue}{CLOSE_SETTING_INJECTION}')
                    completeSettingValue = c.NOTHING
        return settingValueList
    return []

def containsValidSettingInjection(settingValue) :
    if ObjectHelper.isNotNone(settingValue) and StringHelper.isNotBlank(settingValue) and 0 < settingValue.count(OPEN_SETTING_INJECTION) and settingValue.count(c.OPEN_DICTIONARY) == settingValue.count(c.CLOSE_DICTIONARY) :
        splitedSettingValue = settingValue.split(OPEN_SETTING_INJECTION)
        settingValueList = []
        completeSettingValue = c.NOTHING
        for segment in splitedSettingValue if settingValue.startswith(OPEN_SETTING_INJECTION) else splitedSettingValue[1:] :
            if ObjectHelper.isNotNone(segment) :
                if ObjectHelper.isNotNone(segment.count(c.OPEN_DICTIONARY)) and not segment.count(c.OPEN_DICTIONARY) == segment.count(c.CLOSE_DICTIONARY) and 0 < segment.count(c.OPEN_DICTIONARY) :
                    completeSettingValue += segment
                else :
                    splitedSegment = segment.split(CLOSE_SETTING_INJECTION)
                    completeSettingValue += splitedSegment[0]
                    settingValueList.append(f'{OPEN_SETTING_INJECTION}{completeSettingValue}{CLOSE_SETTING_INJECTION}')
                    completeSettingValue = c.NOTHING
        return len(splitedSettingValue) == len(settingValueList) if settingValue.startswith(OPEN_SETTING_INJECTION) else len(splitedSettingValue) == len(settingValueList) + 1
    elif ObjectHelper.isList(settingValue) :
        return containsValidSettingInjection(str(settingValue))
    return False

def isSettingInjection(settingValue) :
    return ObjectHelper.isNotNone(settingValue) and (
        isinstance(settingValue, str) and
        settingValue.startswith(OPEN_SETTING_INJECTION) and
        settingValue.endswith(CLOSE_SETTING_INJECTION) and
        containsOnlyOneSettingInjection(settingValue)
    )

def containsSettingInjection(settingValue) :
    return False if not containsValidSettingInjection(settingValue) else 0 < len(getSettingInjectionListFromSettingValue(settingValue))

def containsOnlyOneSettingInjection(settingValue) :
    return False if not containsValidSettingInjection(settingValue) else 1 == len(getSettingInjectionListFromSettingValue(settingValue))

def getSettingInjectionValueIgnoringFallbackSettingTree(settingKey, settingValue, nodeKey, settingTree, fallbackSettingTree=None) :
    return getSettingInjectionValue(settingKey, settingValue, nodeKey, settingTree)

def getSettingInjectionValue(settingKey, settingValue, nodeKey, settingTree) :
    unwrapedSettingInjectionValue = getUnwrappedSettingInjection(settingValue)
    # print(f'getSettingInjectionValue: settingKey: {settingKey}, unwrapedSettingInjectionValue: {unwrapedSettingInjectionValue}, settingTree: {settingTree}')
    existingValue = None
    try:
        # print(f'nodeKey: {nodeKey}, settingKey: {settingKey}, settingTree: {settingTree}')
        existingValue = safelyAccessTree(nodeKey,settingTree).get(settingKey)
        # print(f'existingValue: {existingValue}')
    except Exception as exception:
        # print(f'nodeKey: {nodeKey}, settingKey: {settingKey}, settingTree: {settingTree}, safelyAccessTree(nodeKey,settingTree): {safelyAccessTree(nodeKey,settingTree)}')
        LogHelper.log(getSettingInjectionValue, 'Error while evaluating existing value', exception)
    if ObjectHelper.isNotEmpty(existingValue) and not isSettingInjection(existingValue) and not containsSettingInjection(existingValue) and not isSettingKey(existingValue):
        # if ObjectHelper.isNotEmpty(existingValue) and not isSettingInjection(existingValue) and not containsSettingInjection(existingValue) and not isSettingKey(existingValue) and not (isinstance(existingValue) and existingValue.startswith(c.COLON)):
        return existingValue
    if isSettingKey(unwrapedSettingInjectionValue) :
        selfReferenceSettingValue = safelyAccessTree(unwrapedSettingInjectionValue, settingTree)
        if ObjectHelper.isNone(selfReferenceSettingValue) :
            raise Exception(f'Not possible to associate "{nodeKey}{c.DOT}{settingKey}" key to "{unwrapedSettingInjectionValue}" value. "{unwrapedSettingInjectionValue}" value is either not defined or not delayed defined. Example: "{OPEN_SETTING_INJECTION}:{nodeKey}{c.DOT}{settingKey}{CLOSE_SETTING_INJECTION}"')
            # return settingValue
        return selfReferenceSettingValue
    environmentKey = unwrapedSettingInjectionValue.split(c.COLON)[0]
    environmentValue = EnvironmentHelper.get(environmentKey)
    # print()
    # print()
    # print(f'getSettingInjectionValue: {settingKey} --> {environmentKey}: getValue(environmentValue): {getValue(environmentValue)}, type: {type(getValue(environmentValue))}') ##remove
    # print(f'getSettingInjectionValue: {settingKey} --> {environmentKey}: environmentValue: {environmentValue}, type: {type(environmentValue)}') ##remove
    if environmentValue :
        return getValue(environmentValue, ignoreCollections=True)
        # return environmentValue
    else :
        return getFilteredSetting(settingKey, unwrapedSettingInjectionValue, nodeKey, settingTree)

def getUnwrappedSettingInjection(settingValue) :
    if isSettingInjection(settingValue) :
        return settingValue[2:-1]
    return settingValue

def keepSearching(keywordQuery,tree,querySet,history=None):
    if ObjectHelper.isDictionary(tree) :
        for key in tree.keys() :
            if StringHelper.isNotBlank(history) :
                newHistory = f'{history}.{key}'
            else :
                newHistory = f'{key}'
            if StringHelper.isNotBlank(keywordQuery) and key == keywordQuery :
                querySet[newHistory] = tree[key]
            keepSearching(keywordQuery,tree[key],querySet,history=newHistory)

def printNodeTree(
        tree,
        depth,
        settingKeyColor=c.NOTHING,
        settingValueColor=c.NOTHING,
        colonColor=c.NOTHING,
        resetColor=c.NOTHING
    ):
    depthSpace = c.NOTHING
    for nodeDeep in range(depth) :
        depthSpace += f'{c.TAB_UNITS * c.SPACE}'
    depth += 1
    for node in list(tree) :
        if ObjectHelper.isDictionary(tree[node]) :
            print(f'{depthSpace}{settingKeyColor}{node}{colonColor}{c.SPACE}{c.COLON}')
            printNodeTree(
                tree[node],
                depth,
                settingKeyColor=settingKeyColor,
                settingValueColor=settingValueColor,
                colonColor=colonColor,
                resetColor=resetColor
            )
        else :
            print(f'{depthSpace}{settingKeyColor}{node}{colonColor}{c.SPACE}{c.COLON_SPACE}{settingValueColor}{tree[node]}{resetColor}')

def getSettingKeyPrompColor(withColors) :
    return c.DARK_MAGENTA if withColors else c.NOTHING

def getSettingValuePrompColor(withColors) :
    return c.DARK_YELLOW if withColors else c.NOTHING

def getSettingColonPrompColor(withColors) :
    return c.BRIGHT_BLACK if withColors else c.NOTHING

def getSettingResetPrompColor(withColors) :
    return c.DEFAULT_COLOR if withColors else c.NOTHING



import inspect, functools
from python_helper.api.src.domain  import Constant as c
from python_helper.api.src.service import LogHelper, ObjectHelper, StringHelper, RandomHelper


MAXIMUN_ARGUMENTS = 20

UNKNOWN_TYPE_NAME = f'{c.UNKNOWN.lower()} type'
UNDEFINED = 'undefined'

def thisIsAFunction():
    ...

METHOD_TYPE_NAMES = (
    c.TYPE_METHOD,
    c.TYPE_BUILTIN_FUNCTION_OR_METHOD
    # , type(thisIsAFunction).__name__ ###- it cannot be a function
)


def hasAttributeOrMethod(instance, name):
    return False if ObjectHelper.isNone(instance) or ObjectHelper.isNone(name) else hasattr(instance, name)


def getAttributeOrMethod(instance, name, muteLogs=False, default=None):
    attributeOrMethodInstance = None
    if ObjectHelper.isNotNone(instance) and ObjectHelper.isNotNone(name):
        try :
            attributeOrMethodInstance = default if not hasattr(instance, name) else getattr(instance, name)
        except Exception as exception :
            if not muteLogs :
                LogHelper.warning(getAttributeOrMethod, f'Not possible to get "{name}" from "{getClassName(instance, typeClass=c.TYPE_CLASS, muteLogs=muteLogs) if ObjectHelper.isNotNone(instance) else instance}" instance', exception=exception)
    return default if ObjectHelper.isNone(attributeOrMethodInstance) else attributeOrMethodInstance


def setAttributeOrMethod(instance, name, attributeOrMethodInstance, muteLogs=False):
    if ObjectHelper.isNotNone(instance) and ObjectHelper.isNotNone(name):
        try :
            setattr(instance, name, attributeOrMethodInstance)
        except Exception as exception :
            if not muteLogs :
                LogHelper.warning(setAttributeOrMethod, f'Not possible to set "{name}:{attributeOrMethodInstance}" to "{getClassName(instance, typeClass=c.TYPE_CLASS, muteLogs=muteLogs) if ObjectHelper.isNotNone(instance) else instance}" instance', exception=exception)


def getAttributeOrMethodByNamePath(instance, namePath, muteLogs=False, default=None):
    if not (ObjectHelper.isEmpty(instance) or ObjectHelper.isEmpty(namePath) or namePath.startswith(c.DOT) or namePath.endswith(c.DOT)):
        return unsafellyGetAttributeOrMethodByNamePath(instance, namePath, muteLogs=muteLogs, default=default)
    return default


def unsafellyGetAttributeOrMethodByNamePath(instance, namePath, muteLogs=False, default=None):
    splittedMethodPath = namePath.split(c.DOT)
    if 1 > len(splittedMethodPath):
        return default if ObjectHelper.isNone(instance) else instance
    elif 1 == len(splittedMethodPath):
        return getAttributeOrMethod(instance, splittedMethodPath[0], muteLogs=muteLogs, default=default)
    else:
        return unsafellyGetAttributeOrMethodByNamePath(
            getAttributeOrMethod(instance, splittedMethodPath[0], muteLogs=muteLogs, default=default),
            StringHelper.join(splittedMethodPath[1:], character=c.DOT),
            muteLogs = muteLogs,
            default = default
        )


def getAttributeOrMethodNameList(instanceClass, muteLogs=False):
    objectNullArgsInstance = instanciateItWithNoArgsConstructor(instanceClass, muteLogs=muteLogs)
    return [
        attributeOrMethodName
        for attributeOrMethodName in dir(objectNullArgsInstance)
        if isNotPrivate(attributeOrMethodName, muteLogs=muteLogs)
    ]


def isAttributeName(attributeName, objectNullArgsInstance, muteLogs=False):
    return isNotPrivate(attributeName, muteLogs=muteLogs) and isNotMethod(objectNullArgsInstance, attributeName, muteLogs=muteLogs)


def getAttributeNameList(instanceClass, muteLogs=False):
    objectNullArgsInstance = instanciateItWithNoArgsConstructor(instanceClass, muteLogs=muteLogs)
    return [
        attributeName
        for attributeName in dir(objectNullArgsInstance)
        if isAttributeName(attributeName, objectNullArgsInstance, muteLogs=muteLogs)
    ]


def getMethodNameList(instanceClass, muteLogs=False):
    objectNullArgsInstance = instanciateItWithNoArgsConstructor(instanceClass, muteLogs=muteLogs)
    return [
        methodName
        for methodName in dir(objectNullArgsInstance)
        if isNotPrivate(methodName, muteLogs=muteLogs) and isMethod(objectNullArgsInstance, methodName, muteLogs=muteLogs)
    ]


def isMethodClass(methodClass, muteLogs=False):
    return False if ObjectHelper.isNone(methodClass) else methodClass.__name__ in METHOD_TYPE_NAMES


def isNotMethodClass(methodClass, muteLogs=False):
    return False if ObjectHelper.isNone(methodClass) else not isMethodClass(methodClass, muteLogs=muteLogs)


def isMethodInstance(methodInstance, muteLogs=False):
    return isMethodClass(type(methodInstance), muteLogs=muteLogs)


def isNotMethodInstance(methodInstance, muteLogs=False):
    return not isMethodInstance(methodInstance, muteLogs=muteLogs)


def isMethod(objectInstance, name, muteLogs=False):
    if ObjectHelper.isNone(objectInstance) or StringHelper.isBlank(name):
        return False
    return isMethodInstance(getAttributeOrMethod(objectInstance, name, muteLogs=muteLogs), muteLogs=muteLogs)


def isNotMethod(objectInstance, name, muteLogs=False):
    if ObjectHelper.isNone(objectInstance) or StringHelper.isBlank(name):
        return False
    return isNotMethodInstance(getAttributeOrMethod(objectInstance, name, muteLogs=muteLogs), muteLogs=muteLogs)


def isNotFunction(methodInstance):
    return not isinstance(methodInstance, isinstance(type(thisIsAFunction)))


def instanciateItWithNoArgsConstructor(targetClass, amountOfNoneArgs=0, args=None, muteLogs=False):
    if ObjectHelper.isNone(args):
        args = []
    for _ in range(amountOfNoneArgs):
        args.append(None)
    objectInstance = None
    for _ in range(MAXIMUN_ARGUMENTS):
        try :
            objectInstance = targetClass(*args)
            break
        except :
            args.append(None)
    if not isinstance(objectInstance, targetClass):
        raise Exception(f'Not possible to instanciate {getClassName(targetClass, typeClass=c.TYPE_CLASS, muteLogs=muteLogs)} with None as args constructor')
    return objectInstance


def getArgsOrder(targetClass, muteLogs=False):
    noneArgs = []
    noneInstance = instanciateItWithNoArgsConstructor(targetClass, amountOfNoneArgs=0, args=noneArgs, muteLogs=muteLogs)
    strArgs = []
    for arg in range(len(noneArgs)):
        strArgs.append(RandomHelper.string(minimum=10))
    try :
        instance = targetClass(*strArgs)
        instanceDataDictionary = getAttributeDataDictionary(instance, muteLogs=muteLogs)
        argsOrderDictionary = {}
        for key,value in instanceDataDictionary.items():
            if StringHelper.isNotBlank(value):
                argsOrderDictionary[strArgs.index(value)] = key
        argsOrder = [argsOrderDictionary[key] for key in sorted(argsOrderDictionary)]
    except Exception as exception :
        errorMessage = f'Not possible to get args order from "{getName(targetClass)}" target class'
        LogHelper.error(getArgsOrder, errorMessage, exception)
        raise Exception(errorMessage)
    return argsOrder


def isNotPrivate(attributeOrMethodName, muteLogs=False):
    return StringHelper.isNotBlank(attributeOrMethodName) and (
        not attributeOrMethodName.startswith(f'{2 * c.UNDERSCORE}') and
        not attributeOrMethodName.startswith(c.UNDERSCORE) and
        not ObjectHelper.METADATA_NAME == attributeOrMethodName
    )


def getAttributePointerList(instance, muteLogs=False):
    return [
        getattr(instance, instanceAttributeOrMethodName)
        for instanceAttributeOrMethodName in dir(instance)
        if isNotPrivate(instanceAttributeOrMethodName, muteLogs=muteLogs)
    ]


def getAttributeDataList(instance, muteLogs=False):
    return [
        (getattr(instance, instanceAttributeName), instanceAttributeName)
        for instanceAttributeName in dir(instance)
        if isAttributeName(instanceAttributeName, instance, muteLogs=muteLogs)
    ]


def getAttributeDataDictionary(instance, muteLogs=False):
    '''It can be a function, but not a method...'''
    instanceDataDictionary = {}
    for name in dir(instance):
        if isAttributeName(name, instance, muteLogs=muteLogs):
            instanceDataDictionary[name] = getattr(instance, name)
    return instanceDataDictionary


def overrideSignatures(toOverride, original, forceName=None, forceModuleName=None):
    try :
        if ObjectHelper.isNotNone(original):
            toOverride.__name__ = original.__name__ if ObjectHelper.isNone(forceName) else set(forceName)
            toOverride.__qualname__ = original.__qualname__ if ObjectHelper.isNone(forceName) else set(forceName)
            toOverride.__module__ = original.__module__ if ObjectHelper.isNone(forceName) else set(c.NOTHING)
        else :
            toOverride.__name__ = forceName if ObjectHelper.isNotNone(forceName) else set(toOverride.__name__)
            toOverride.__qualname__ = forceName if ObjectHelper.isNotNone(forceName) else set(toOverride.__qualname__)
            toOverride.__module__ = forceModuleName if ObjectHelper.isNotNone(forceModuleName) else set(toOverride.__module__)
    except Exception as exception :
        LogHelper.error(overrideSignatures, f'''Not possible to override signatures of {toOverride} by signatures of {original} method''', exception)
        raise exception


def getClass(thing, typeClass=None, muteLogs=False):
    thingClass = None
    try :
        if ObjectHelper.isEmpty(thing):
            thingClass = typeClass
        else :
            thingClass = thing.__class__
    except Exception as exception :
        thingClass = type(None)
        if not muteLogs :
            LogHelper.warning(getClass, f'Not possible to get class of {thing}. Returning {thingClass} insted', exception=exception)
    return thingClass


def getName(thing, typeName=None, muteLogs=False):
    name = None
    try :
        if ObjectHelper.isEmpty(thing):
            name = getUndefindeName(typeName)
        else :
            name = thing.__name__
    except Exception as exception :
        name = getUndefindeName(typeName)
        if not muteLogs :
            LogHelper.warning(getName, f'Not possible to get name of {thing}. Returning {name} insted', exception=exception)
    return name


def getClassName(thing, typeClass=None, muteLogs=False):
    name = None
    try :
        if ObjectHelper.isEmpty(thing):
            name = getUndefindeName(typeClass)
        else :
            name = getName(getClass(thing, muteLogs=muteLogs), muteLogs=muteLogs)
    except Exception as exception :
        name = getUndefindeName(typeClass)
        if not muteLogs :
            LogHelper.warning(getClassName, f'Not possible to get class name of {thing}. Returning {name} insted', exception=exception)
    return name


def getMethodClassName(instanceClass):
    return instanceClass.__qualname__.split(c.DOT)[0]


def getModuleName(thing, typeModule=None, muteLogs=False):
    name = None
    try :
        if ObjectHelper.isEmpty(thing):
            name = getUndefindeName(typeModule)
        else :
            name = thing.__module__.split(c.DOT)[-1]
    except Exception as exception :
        name = getUndefindeName(typeModule)
        if not muteLogs :
            LogHelper.warning(getModuleName, f'Not possible to get module name of {thing}. Returning {name} insted', exception=exception)
    return name


def getMethodModuleNameDotName(instance):
    return f'{getModuleName(instance)}{c.DOT}{getName(instance)}'


def getUndefindeName(typeThing):
    if ObjectHelper.isEmpty(typeThing):
        return f'({UNDEFINED})'
    else :
        return f'({typeThing} {UNDEFINED})'


def getParentClass(instance):
    instanceParent = None
    try :
        instanceParent = unsafelyGetInstanceParent(instance)
    except Exception as exception:
        LogHelper.wrapper(getInstanceParent, 'Failed to get instance parent', exception)
    return instanceParent


def unsafelyGetInstanceParent(instance):
    if isinstance(instance, functools.partial):
        return unsafelyGetInstanceParent(instance.func)
    if inspect.ismethod(instance) or (inspect.isbuiltin(instance) and getattr(instance, '__self__', None) is not None and getattr(instance.__self__, '__class__', None)):
        for cls in inspect.getmro(instance.__self__.__class__):
            if instance.__name__ in cls.__dict__:
                return cls
        instance = getattr(instance, '__func__', instance)
    if inspect.isfunction(instance):
        cls = getattr(inspect.getmodule(instance), instance.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(instance, '__objclass__', None)


def getItNaked(it):
    printDetails(it)
    printClass(it)
    try :
        LogHelper.prettyPython(getAttributePointerList, 'getAttributePointerList', getAttributePointerList(it), logLevel=LogHelper.DEBUG)
    except : pass
    try :
        LogHelper.prettyPython(getAttributeAndMethodNameList, 'getAttributeAndMethodNameList', getAttributeAndMethodNameList(it), logLevel=LogHelper.DEBUG)
    except : pass
    try :
        LogHelper.prettyPython(getAttributeNameList, 'getAttributeNameList', getAttributeNameList(it), logLevel=LogHelper.DEBUG)
    except : pass
    try :
        LogHelper.prettyPython(getAttributeDataList, 'getAttributeDataList', getAttributeDataList(it), logLevel=LogHelper.DEBUG)
    except : pass
    try :
        LogHelper.prettyPython(getAttributeDataDictionary, 'getAttributeDataDictionary', getAttributeDataDictionary(it), logLevel=LogHelper.DEBUG)
    except : pass


def printDetails(toDetail):
    print(f'{2 * c.TAB}printDetails({toDetail}):')
    try :
        print(f'{2 * c.TAB}type({toDetail}).__name__ = {getName(type(toDetail), typeName=UNKNOWN_TYPE_NAME)}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}type({toDetail}).__class__ = {type(toDetail).__class__}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}type({toDetail}).__class__.__module__ = {type(toDetail).__class__.__module__}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}type({toDetail}).__class__.__name__ = {getName(type(toDetail).__class__, typeName=UNKNOWN_TYPE_NAME)}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}{toDetail}.__class__.__name__ = {getName(toDetail.__class__, typeName=UNKNOWN_TYPE_NAME)}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}{toDetail}.__class__.__module__ = {toDetail.__class__.__module__}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}{toDetail}.__class__.__qualname__ = {toDetail.__class__.__qualname__}')
    except :
        pass


def printClass(instanceClass):
    print(f'{2 * c.TAB}printClass({instanceClass}):')
    try :
        print(f'{2 * c.TAB}{instanceClass}.__name__ = {getName(instanceClass, typeName=UNKNOWN_TYPE_NAME)}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}{instanceClass}.__module__ = {instanceClass.__module__}')
    except :
        pass
    try :
        print(f'{2 * c.TAB}{instanceClass}.__qualname__ = {instanceClass.__qualname__}')
    except :
        pass



from python_helper import TestHelper
TestHelper.run(__file__, inspectGlobals=False)
# TestHelper.run(
#     __file__,
#     runOnly = [
#         # 'ReflectionHelperTest.instanciateItWithNoArgsConstructor_withSuccess',
#         'ReflectionHelperTest.getAttributeOrMethodByNamePath'
#     ]
# )
# TestHelper.run(__file__,
#     testStatus = True,
#     logStatus = True,
#     logResult = True,
#     inspectGlobals=False
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         # 'ObjectHelperTest.equal_whenListOfDictionaries',
#         'ObjectHelperTest.equal_whenDictionary'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SettingHelperTest.updateActiveEnvironment_withSuccess',
#         'SettingHelperTest.mustReadSettingFile',
#         'SettingHelperTest.mustNotReadSettingFile',
#         'SettingHelperTest.mustPrintSettingTree',
#         'SettingHelperTest.mustHandleSettingValueInFallbackSettingTree',
#         'SettingHelperTest.mustHandleSettingValueInFallbackSettingTree_whenFallbackSettingFilePathIsPassedInsted',
#         'SettingHelperTest.getSettingTree_whenIsSettingKeyActuallyContainsSettingKey',
#         'SettingHelperTest.getSettingTree_fallbackPriority',
#         'SettingHelperTest.getSettingTree_otherApplication',
#         'SettingHelperTest.getSettingTree_whenThereAreNoneValuesAllOverThePlace',
#         'SettingHelperTest.getBooleanSetting'
#     ],
#     # testStatus = False,
#     # logStatus = False,
#     # logResult = True,
#     # inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 2,
#     testStatus = False,
#     logStatus = True,
#     logResult = True,
#     inspectGlobals = True
# )
# TestHelper.run(
#     __file__,
#     # times = 10,
#     runOnly = [
#         'LogHelperTest.mustLogWithColors',
#         'LogHelperTest.mustLogWithoutColors',
#         'LogHelperTest.mustLogWithoutColorsAsWell',
#         'LogHelperTest.mustLogEnvironmentSettings',
#         'LogHelperTest.mustLogPretyPythonWithoutColors',
#         'LogHelperTest.mustLogPretyPythonWithColors',
#         'LogHelperTest.mustLogPretyJsonWithColors',
#         'LogHelperTest.mustPrintMessageLog_withColors'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = False,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 100,
#     runOnly = [
#         'StringHelperTest.prettyJson_withSucces'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = True,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 100,
#     runOnly = [
#         'StringHelperTest.prettyPython_withSucces'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = True,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 10,
#     testStatus = False,
#     logStatus = False,
#     logResult = True,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 1,
#     runOnly = [
#         'StringHelperTest.prettifyPerformance'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = True,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 1,
#     runOnly = [
#         'ModuleImportsTest.allImportedSuccesfuly'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = False,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 1,
#     runOnly = [
#         'SettingHelperTest.getSettingTree_fallbackPriority'
#     ],
#     settingStatus = False,
#     testStatus = False,
#     logStatus = False,
#     logResult = False,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 1,
#     runOnly = [
#         'MethodAnnotationTest.Function_withSuccess'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = False,
#     inspectGlobals = False
# )
# TestHelper.run(
#     __file__,
#     times = 1,
#     runOnly = [
#         'LogHelperTest.mustLogWithColors'
#     ],
#     testStatus = False,
#     logStatus = False,
#     logResult = False,
#     inspectGlobals = True
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'ReflectionHelperTest.getName_manyCases'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'EnvironmentHelperTest.osIdentifierTest'
#     ]
# )
# TestHelper.run(
#     __file__,
#     runOnly = [
#         'SettingHelperTest.getSettingTree_whenThereAreNoneValuesAllOverThePlace'
#     ]
# )



from python_helper.api.src.service import ObjectHelper, ReflectionHelper, LogHelper

def generatorInstance() :
    while True :
        yield False
        break

def leftEqual(left, right, visitedIdInstances, muteLogs=True) :
    if ObjectHelper.isNone(left) or ObjectHelper.isNone(right) :
        return left is None and right is None
    isEqual = True
    leftIsCollection = ObjectHelper.isCollection(left)
    rightIsCollection = ObjectHelper.isCollection(right)
    if leftIsCollection and rightIsCollection :
        if len(left) == len(right) :
            for itemLeft, itemRight in zip(left, right) :
                if isEqual :
                    isEqual = isEqual and ObjectHelper.equals(itemLeft, itemRight, visitedIdInstances=visitedIdInstances, muteLogs=muteLogs)
                else :
                    break
            return isEqual
        else :
            return False
    elif (leftIsCollection and not rightIsCollection) or (not leftIsCollection and rightIsCollection) :
        return False
    else :
        attrinuteDataList = ReflectionHelper.getAttributeDataList(left)
        if not muteLogs :
            LogHelper.prettyPython(leftEqual, f'{left} data list', attrinuteDataList, logLevel=LogHelper.DEBUG, condition=not muteLogs)
        if 0 == len(attrinuteDataList) :
            return False
        for value, name in attrinuteDataList :
            if isEqual :
                isEqual = isEqual and ObjectHelper.equals(value, ReflectionHelper.getAttributeOrMethod(right, name), visitedIdInstances=visitedIdInstances, muteLogs=muteLogs)
            else :
                break
        return isEqual
